"""Tests for metadata."""

import collections
import datetime
import os
import unittest

from aspecd import metadata, exceptions, utils


class TestPhysicalQuantity(unittest.TestCase):
    def setUp(self):
        self.physical_quantity = metadata.PhysicalQuantity()

    def test_instantiate_class(self):
        pass

    def test_has_value_property(self):
        self.assertTrue(hasattr(self.physical_quantity, "value"))

    def test_value_is_float(self):
        self.assertTrue(isinstance(self.physical_quantity.value, type(0.0)))

    def test_set_wrong_type_for_value_fails(self):
        with self.assertRaises(TypeError):
            self.physical_quantity.value = 0

    def test_set_value(self):
        value = 5.0
        self.physical_quantity.value = value
        self.assertEqual(self.physical_quantity.value, value)

    def test_has_unit_property(self):
        self.assertTrue(hasattr(self.physical_quantity, "unit"))

    def test_has_dimension_property(self):
        self.assertTrue(hasattr(self.physical_quantity, "dimension"))

    def test_has_name_property(self):
        self.assertTrue(hasattr(self.physical_quantity, "name"))

    def test_instantiate_with_string(self):
        physical_quantity = metadata.PhysicalQuantity("5 m")
        self.assertEqual(physical_quantity.value, 5.0)
        self.assertEqual(physical_quantity.unit, "m")

    def test_instantiate_with_string_without_unit(self):
        physical_quantity = metadata.PhysicalQuantity("5")
        self.assertEqual(physical_quantity.value, 5.0)
        self.assertFalse(physical_quantity.unit)

    def test_instantiate_with_value_unit(self):
        physical_quantity = metadata.PhysicalQuantity(value=5.0, unit="m")
        self.assertEqual(physical_quantity.value, 5.0)
        self.assertEqual(physical_quantity.unit, "m")

    def test_string_representation(self):
        string = "5.0 m"
        physical_quantity = metadata.PhysicalQuantity(string)
        self.assertEqual(str(physical_quantity), string)

    def test_string_representation_with_empty_value_unit(self):
        self.assertEqual(str(self.physical_quantity), "")

    def test_string_representation_when_value_is_zero(self):
        string = "0.0 m"
        physical_quantity = metadata.PhysicalQuantity(string)
        self.assertEqual(str(physical_quantity), string)

    def test_commensurable_with_same_unit(self):
        string = "5.0 m"
        physical_quantity1 = metadata.PhysicalQuantity(string)
        physical_quantity2 = metadata.PhysicalQuantity(string)
        self.assertTrue(physical_quantity1.commensurable(physical_quantity2))

    def test_commensurable_with_same_dimension(self):
        dimension = "T"
        physical_quantity1 = metadata.PhysicalQuantity()
        physical_quantity1.dimension = dimension
        physical_quantity2 = metadata.PhysicalQuantity()
        physical_quantity2.dimension = dimension
        self.assertTrue(physical_quantity1.commensurable(physical_quantity2))

    def test_incommensurable_with_different_units(self):
        physical_quantity1 = metadata.PhysicalQuantity("5.0 m")
        physical_quantity2 = metadata.PhysicalQuantity("3 s")
        self.assertFalse(physical_quantity1.commensurable(physical_quantity2))

    def test_incommensurable_with_different_dimensions(self):
        physical_quantity1 = metadata.PhysicalQuantity()
        physical_quantity1.dimension = "T"
        physical_quantity2 = metadata.PhysicalQuantity()
        physical_quantity2.dimension = "L"
        self.assertFalse(physical_quantity1.commensurable(physical_quantity2))

    def test_incommensurable_if_has_no_unit(self):
        physical_quantity = metadata.PhysicalQuantity("5.0 m")
        self.assertFalse(physical_quantity.commensurable([]))

    def test_incommensurable_if_has_no_dimension(self):
        physical_quantity = metadata.PhysicalQuantity()
        physical_quantity.dimension = "T"
        self.assertFalse(physical_quantity.commensurable([]))

    def test_set_value_and_unit_from_string(self):
        self.physical_quantity.from_string("5 m")
        self.assertEqual(self.physical_quantity.value, 5.0)
        self.assertEqual(self.physical_quantity.unit, "m")

    def test_set_value_and_unit_from_empty_string(self):
        self.physical_quantity.from_string("")
        self.assertFalse(self.physical_quantity.value)
        self.assertFalse(self.physical_quantity.unit)

    def test_set_value_and_unit_from_na_string(self):
        self.physical_quantity.from_string("N/A")
        self.assertFalse(self.physical_quantity.value)
        self.assertFalse(self.physical_quantity.unit)

    def test_set_value_and_unit_from_na_string_case_insensitive(self):
        self.physical_quantity.from_string("n/a")
        self.assertFalse(self.physical_quantity.value)
        self.assertFalse(self.physical_quantity.unit)

    def test_empty_string_clears_value_and_unit(self):
        self.physical_quantity.from_string("5 m")
        self.physical_quantity.from_string("")
        self.assertFalse(self.physical_quantity.value)
        self.assertFalse(self.physical_quantity.unit)

    def test_to_dict(self):
        self.physical_quantity.from_string("5 m")
        test_dict = {"value": 5.0, "unit": "m", "dimension": "", "name": ""}
        to_dict = self.physical_quantity.to_dict()
        self.assertDictEqual(test_dict, to_dict)

    def test_from_dict(self):
        dict_ = self.physical_quantity.to_dict()
        dict_["name"] = "foo"
        self.physical_quantity.from_dict(dict_)
        self.assertEqual(dict_["name"], self.physical_quantity.name)


class TestMetadata(unittest.TestCase):
    def setUp(self):
        self.metadata = metadata.Metadata()

    def test_instantiate_class(self):
        pass

    def test_set_properties_from_dict(self):
        dict_ = {
            "purpose": "Kill time",
            "operator": "John Doe",
            "labbook": "loi:42.1001/foo/bar",
        }
        for key in dict_:
            setattr(self.metadata, key, "")
        self.metadata.from_dict(dict_)
        for key in dict_:
            self.assertEqual(getattr(self.metadata, key), dict_[key])

    def test_set_properties_from_dict_case_insensitive(self):
        dict_ = {
            "Purpose": "Kill time",
            "Operator": "John Doe",
            "Labbook": "loi:42.1001/foo/bar",
        }
        for key in dict_:
            setattr(self.metadata, key.lower(), "")
        self.metadata.from_dict(dict_)
        for key in dict_:
            self.assertEqual(getattr(self.metadata, key.lower()), dict_[key])

    def test_set_properties_from_dict_with_spaces_in_keys(self):
        dict_ = {
            "Purpose": "Kill time",
            "Operator": "John Doe",
            "Lab book": "loi:42.1001/foo/bar",
        }
        for key in dict_:
            setattr(self.metadata, key.replace(" ", "_").lower(), "")
        self.metadata.from_dict(dict_)
        for key in dict_:
            self.assertEqual(
                dict_[key],
                getattr(self.metadata, key.replace(" ", "_").lower()),
            )

    def test_instantiate_properties_from_dict_with_physical_quantities(self):
        dict_ = {
            "purpose": "Kill time",
            "operator": "John Doe",
            "labbook": "loi:42.1001/foo/bar",
            "length": "1.0 mm",
        }
        metadata_class = metadata.Metadata
        for key in dict_:
            setattr(metadata_class, key, "")
        metadata_class.length = metadata.PhysicalQuantity()
        metadata_ = metadata_class(dict_)
        for key in dict_:
            if isinstance(getattr(metadata_, key), metadata.PhysicalQuantity):
                self.assertEqual(str(getattr(metadata_, key)), dict_[key])
            else:
                self.assertEqual(getattr(metadata_, key), dict_[key])

    def test_set_properties_from_dict_with_physical_quantities(self):
        dict_ = {
            "purpose": "Kill time",
            "operator": "John Doe",
            "labbook": "loi:42.1001/foo/bar",
            "length": "1.0 mm",
        }
        for key in dict_:
            setattr(self.metadata, key, "")
        self.metadata.length = metadata.PhysicalQuantity()
        self.metadata.from_dict(dict_)
        for key in dict_:
            if isinstance(
                getattr(self.metadata, key), metadata.PhysicalQuantity
            ):
                self.assertEqual(str(getattr(self.metadata, key)), dict_[key])
            else:
                self.assertEqual(getattr(self.metadata, key), dict_[key])

    def test_to_dict(self):
        dict_ = {
            "purpose": "Kill time",
            "operator": "John Doe",
            "labbook": "loi:42.1001/foo/bar",
        }
        for key in dict_:
            setattr(self.metadata, key, "")
        self.metadata.from_dict(dict_)
        to_dict = self.metadata.to_dict()
        self.assertDictEqual(dict_, to_dict)

    def test_to_dict_with_physical_quantities(self):
        dict_ = {
            "purpose": "Kill time",
            "operator": "John Doe",
            "labbook": "loi:42.1001/foo/bar",
            "length": "1.0 mm",
        }
        for key in dict_:
            setattr(self.metadata, key, "")
        self.metadata.length = metadata.PhysicalQuantity()
        self.metadata.from_dict(dict_)
        test_dict = {
            "purpose": "Kill time",
            "operator": "John Doe",
            "labbook": "loi:42.1001/foo/bar",
            "length": {
                "value": 1.0,
                "unit": "mm",
                "dimension": "",
                "name": "",
            },
        }
        to_dict = self.metadata.to_dict()
        self.assertDictEqual(test_dict, to_dict)

    def test_to_dict_preserves_order(self):
        arguments = ["purpose", "operator", "labbook"]
        for argument in arguments:
            setattr(self.metadata, argument, "")
        to_dict = self.metadata.to_dict()
        self.assertEqual(arguments, list(to_dict.keys()))


class TestTemperatureControl(unittest.TestCase):
    def setUp(self):
        self.temperature_control = metadata.TemperatureControl()

    def test_instantiate_class(self):
        pass

    def test_has_controlled_property(self):
        self.assertTrue(hasattr(self.temperature_control, "controlled"))

    def test_has_temperature_property(self):
        self.assertTrue(hasattr(self.temperature_control, "temperature"))

    def test_has_controller_property(self):
        self.assertTrue(hasattr(self.temperature_control, "controller"))

    def test_instantiate_with_temperature_from_string(self):
        temperature_control = metadata.TemperatureControl(temperature="270 K")
        self.assertEqual(temperature_control.temperature.value, 270.0)
        self.assertEqual(temperature_control.temperature.unit, "K")

    def test_instantiate_with_temperature_sets_controlled(self):
        temperature_control = metadata.TemperatureControl(temperature="270 K")
        self.assertTrue(temperature_control.controlled)

    def test_setting_temperature_sets_controlled(self):
        self.temperature_control.temperature.from_string("270 K")
        self.assertTrue(self.temperature_control.controlled)

    def test_instantiate_properties_from_dict(self):
        dict_ = {
            "controlled": True,
            "temperature": "270 K",
            "controller": "Oxford ITC 503S",
        }
        temperature_control = metadata.TemperatureControl(dict_)
        self.assertTrue(temperature_control.controlled)
        self.assertEqual(temperature_control.temperature.value, 270.0)
        self.assertEqual(temperature_control.temperature.unit, "K")
        self.assertEqual(temperature_control.controller, dict_["controller"])

    def test_instantiate_properties_from_dict_overwrites_controlled(self):
        dict_ = {
            "controlled": False,
            "temperature": "270 K",
            "controller": "Oxford ITC 503S",
        }
        temperature_control = metadata.TemperatureControl(dict_)
        self.assertFalse(temperature_control.controlled)

    def test_set_properties_from_dict(self):
        dict_ = {
            "controlled": True,
            "temperature": "270 K",
            "controller": "Oxford ITC 503S",
        }
        self.temperature_control.from_dict(dict_)
        self.assertTrue(self.temperature_control.controlled)
        self.assertEqual(self.temperature_control.temperature.value, 270.0)
        self.assertEqual(self.temperature_control.temperature.unit, "K")
        self.assertEqual(
            self.temperature_control.controller, dict_["controller"]
        )

    def test_set_properties_from_dict_with_missing_controlled_field(self):
        dict_ = {"temperature": "270 K", "controller": "Oxford ITC 503S"}
        self.temperature_control.from_dict(dict_)
        self.assertTrue(self.temperature_control.controlled)
        self.assertEqual(self.temperature_control.temperature.value, 270.0)
        self.assertEqual(self.temperature_control.temperature.unit, "K")
        self.assertEqual(
            self.temperature_control.controller, dict_["controller"]
        )

    def test_set_properties_from_dict_with_missing_temperature_field(self):
        dict_ = {"controller": "Oxford ITC 503S"}
        self.temperature_control.from_dict(dict_)
        self.assertEqual(
            self.temperature_control.controller, dict_["controller"]
        )


class TestMeasurement(unittest.TestCase):
    def setUp(self):
        self.measurement = metadata.Measurement()

    def test_instantiate_class(self):
        pass

    def test_instantiate_properties_from_dict(self):
        dict_ = {
            "purpose": "Kill time",
            "operator": "John Doe",
            "labbook_entry": "loi:42.1001/foo/bar",
        }
        measurement = metadata.Measurement(dict_)
        for key in dict_:
            self.assertEqual(getattr(measurement, key), dict_[key])

    def test_set_properties_from_dict(self):
        dict_ = {
            "purpose": "Kill time",
            "operator": "John Doe",
            "labbook_entry": "loi:42.1001/foo/bar",
        }
        self.measurement.from_dict(dict_)
        for key in dict_:
            self.assertEqual(getattr(self.measurement, key), dict_[key])

    def test_instantiate_start_end_from_dict(self):
        dict_ = {
            "start": {"date": "2017-01-02", "time": "11:00:00"},
            "end": {"date": "2017-01-02", "time": "11:01:00"},
        }
        measurement = metadata.Measurement(dict_)
        fmt = "%Y%m%dT%H%M%S"
        self.assertEqual(measurement.start.strftime(fmt), "20170102T110000")
        self.assertEqual(measurement.end.strftime(fmt), "20170102T110100")

    def test_instantiate_start_end_from_dict_with_datetime(self):
        dict_ = {
            "start": {"date": datetime.date(2017, 1, 2), "time": "11:00:00"},
            "end": {"date": "2017-01-02", "time": "11:01:00"},
        }
        measurement = metadata.Measurement(dict_)
        fmt = "%Y%m%dT%H%M%S"
        self.assertEqual(measurement.start.strftime(fmt), "20170102T110000")
        self.assertEqual(measurement.end.strftime(fmt), "20170102T110100")

    def test_set_start_end_from_dict_with_strings(self):
        dict_ = {"start": "2017-01-02 11:00:00", "end": "2017-01-02 11:01:00"}
        measurement = metadata.Measurement(dict_)
        fmt = "%Y%m%dT%H%M%S"
        self.assertEqual(measurement.start.strftime(fmt), "20170102T110000")
        self.assertEqual(measurement.end.strftime(fmt), "20170102T110100")

    def test_to_dict_preserves_order(self):
        arguments = ["start", "end", "purpose", "operator", "labbook_entry"]
        to_dict = self.measurement.to_dict()
        self.assertEqual(arguments, list(to_dict.keys()))


class TestSample(unittest.TestCase):
    def setUp(self):
        self.sample = metadata.Sample()

    def test_instantiate_class(self):
        pass

    def test_instantiate_properties_from_dict(self):
        dict_ = {"name": "Sample1", "id": 42, "loi": "loi:42.1001/foo/bar"}
        sample = metadata.Sample(dict_)
        for key in dict_:
            self.assertEqual(getattr(sample, key), dict_[key])

    def test_set_properties_from_dict(self):
        dict_ = {"name": "Sample1", "id": 42, "loi": "loi:42.1001/foo/bar"}
        self.sample.from_dict(dict_)
        for key in dict_:
            self.assertEqual(getattr(self.sample, key), dict_[key])


class TestCalculation(unittest.TestCase):
    def setUp(self):
        self.calculation = metadata.Calculation()

    def test_instantiate_class(self):
        pass

    def test_instantiate_properties_from_dict(self):
        dict_ = {"type": "foo", "parameters": {"foo": "bar"}}
        calculation = metadata.Calculation(dict_)
        for key in dict_:
            self.assertEqual(getattr(calculation, key), dict_[key])

    def test_set_properties_from_dict(self):
        dict_ = {"type": "foo", "parameters": {"foo": "bar"}}
        self.calculation.from_dict(dict_)
        for key in dict_:
            self.assertEqual(getattr(self.calculation, key), dict_[key])


class TestDevice(unittest.TestCase):
    def setUp(self):
        self.device = metadata.Device()

    def test_instantiate_class(self):
        pass

    def test_instantiate_properties_from_dict(self):
        dict_ = {"label": "foo"}
        device = metadata.Device(dict_)
        for key in dict_:
            self.assertEqual(getattr(device, key), dict_[key])

    def test_set_properties_from_dict(self):
        dict_ = {"label": "foo"}
        self.device.from_dict(dict_)
        for key in dict_:
            self.assertEqual(getattr(self.device, key), dict_[key])


class TestExperimentalDatasetMetadata(unittest.TestCase):
    def setUp(self):
        self.dataset_metadata = metadata.ExperimentalDatasetMetadata()

    def test_instantiate_class(self):
        pass

    def test_set_measurement_metadata_from_dict(self):
        dict_ = {
            "measurement": {
                "purpose": "Kill time",
                "operator": "John Doe",
                "labbook_entry": "loi:42.1001/foo/bar",
            }
        }
        self.dataset_metadata.from_dict(dict_)
        for key in dict_["measurement"]:
            self.assertEqual(
                dict_["measurement"][key],
                getattr(self.dataset_metadata.measurement, key),
            )

    def test_set_measurement_metadata_from_dict_case_insensitive(self):
        dict_ = {
            "Measurement": {
                "purpose": "Kill time",
                "operator": "John Doe",
                "labbook_entry": "loi:42.1001/foo/bar",
            }
        }
        self.dataset_metadata.from_dict(dict_)
        for key in dict_["Measurement"]:
            self.assertEqual(
                dict_["Measurement"][key],
                getattr(self.dataset_metadata.measurement, key),
            )

    def test_set_properties_from_dict_with_spaces_in_keys(self):
        dict_ = {"Temperature Control": {"controller": "Oxford ITC503S"}}
        self.dataset_metadata.from_dict(dict_)
        for key in dict_["Temperature Control"]:
            self.assertEqual(
                dict_["Temperature Control"][key],
                getattr(self.dataset_metadata.temperature_control, key),
            )

    def test_set_temperature_metadata_from_dict(self):
        dict_ = {
            "temperature_control": {
                "controlled": True,
                "temperature": "270 K",
                "controller": "Oxford ITC 503S",
            }
        }
        self.dataset_metadata.from_dict(dict_)
        self.assertTrue(self.dataset_metadata.temperature_control.controlled)
        self.assertEqual(
            self.dataset_metadata.temperature_control.temperature.value, 270.0
        )
        self.assertEqual(
            self.dataset_metadata.temperature_control.temperature.unit, "K"
        )
        self.assertEqual(
            dict_["temperature_control"]["controller"],
            self.dataset_metadata.temperature_control.controller,
        )

    def test_set_sample_metadata_from_dict(self):
        dict_ = {
            "sample": {
                "name": "Sample1",
                "id": 42,
                "loi": "loi:42.1001/foo/bar",
            }
        }
        self.dataset_metadata.from_dict(dict_)
        for key in dict_["sample"]:
            self.assertEqual(
                dict_["sample"][key],
                getattr(self.dataset_metadata.sample, key),
            )

    def test_to_dict(self):
        dict_ = {
            "sample": {
                "name": "Sample1",
                "id": 42,
                "loi": "loi:42.1001/foo/bar",
            }
        }
        self.dataset_metadata.from_dict(dict_)
        to_dict = self.dataset_metadata.to_dict()
        self.assertDictEqual(dict_["sample"], to_dict["sample"])


class TestCalculatedDatasetMetadata(unittest.TestCase):
    def setUp(self):
        self.dataset_metadata = metadata.CalculatedDatasetMetadata()

    def test_instantiate_class(self):
        pass

    def test_has_calculation_property(self):
        self.assertTrue(hasattr(self.dataset_metadata, "calculation"))

    def test_calculation_property_is_calculation(self):
        self.assertTrue(
            isinstance(
                self.dataset_metadata.calculation, metadata.Calculation
            )
        )


class TestMetadataMapper(unittest.TestCase):
    def setUp(self):
        self.metadata_mapper = metadata.MetadataMapper()
        self.filename = "test_metadata_mappings.yaml"
        self.root_path = os.path.join(
            os.path.split(os.path.split(os.path.abspath(__file__))[0])[0],
            "aspecd",
        )

    def tearDown(self):
        if os.path.exists(os.path.join(self.root_path, self.filename)):
            os.remove(os.path.join(self.root_path, self.filename))

    def write_metadata_mapper_file(self, dict_=None):
        yaml_file = utils.Yaml()
        yaml_file.dict = dict_
        yaml_file.write_to(os.path.join(self.root_path, self.filename))

    def test_instantiate_class(self):
        pass

    def test_has_method_rename_key(self):
        self.assertTrue(callable(self.metadata_mapper.rename_key))

    def test_rename_key(self):
        new_key = "foo"
        old_key = "bar"
        self.metadata_mapper.metadata[old_key] = "bar"
        self.metadata_mapper.rename_key(old_key, new_key)
        self.assertTrue(new_key in self.metadata_mapper.metadata.keys())
        self.assertFalse(old_key in self.metadata_mapper.metadata.keys())

    def test_rename_uppercase_key(self):
        new_key = "foo"
        old_key = "BAR"
        self.metadata_mapper.metadata[old_key] = "bar"
        self.metadata_mapper.rename_key(old_key, new_key)
        self.assertTrue(new_key in self.metadata_mapper.metadata.keys())
        self.assertFalse(old_key in self.metadata_mapper.metadata.keys())

    def test_rename_key_with_not_existing_key_ignores(self):
        new_key = "foobar"
        old_key = "baz"
        self.metadata_mapper.metadata = {"foo": "baz", "bar": {}}
        self.metadata_mapper.rename_key(old_key, new_key)

    def test_rename_key_with_not_existing_key_logs(self):
        new_key = "foobar"
        old_key = "baz"
        self.metadata_mapper.metadata = {"foo": "baz", "bar": {}}
        with self.assertLogs(__package__, level="DEBUG") as cm:
            self.metadata_mapper.rename_key(old_key, new_key)
        self.assertIn('Key "{}" not found'.format(old_key), cm.output[0])

    def test_has_method_combine_items(self):
        self.assertTrue(callable(self.metadata_mapper.combine_items))

    def test_combine_items(self):
        old_keys = (
            collections.OrderedDict()
        )  # ordered dict for reproducibility
        old_keys["key1"] = "bla"
        old_keys["key2"] = "blub"
        new_key = "new_key"
        for key in old_keys.keys():
            self.metadata_mapper.metadata[key] = old_keys[key]
        self.metadata_mapper.combine_items(old_keys, new_key)

        self.assertTrue(new_key in self.metadata_mapper.metadata.keys())
        for key in old_keys.keys():
            self.assertFalse(key in self.metadata_mapper.metadata.keys())
        self.assertEqual(self.metadata_mapper.metadata[new_key], "blablub")

    def test_combine_items_joint_by_pattern(self):
        old_keys = (
            collections.OrderedDict()
        )  # ordered dict for reproducibility
        old_keys["key1"] = "bla"
        old_keys["key2"] = "blub"
        new_key = "new_key"
        for key in old_keys.keys():
            self.metadata_mapper.metadata[key] = old_keys[key]
        self.metadata_mapper.combine_items(old_keys, new_key, pattern=" ")

        self.assertTrue(new_key in self.metadata_mapper.metadata.keys())
        for key in old_keys.keys():
            self.assertFalse(key in self.metadata_mapper.metadata.keys())
        self.assertEqual(self.metadata_mapper.metadata[new_key], "bla blub")

    def test_has_method_keys_to_variable_names(self):
        self.assertTrue(callable(self.metadata_mapper.keys_to_variable_names))

    def test_keys_to_variable_names(self):
        dict_ = {
            "Temperature Control": {
                "Temperature": "278 K",
                "Transfer Line": "Oxford",
            }
        }
        converted = {
            "temperature_control": {
                "temperature": "278 K",
                "transfer_line": "Oxford",
            }
        }
        self.metadata_mapper.metadata = dict_
        self.metadata_mapper.keys_to_variable_names()
        self.assertDictEqual(self.metadata_mapper.metadata, converted)

    def test_has_attribute_mapping(self):
        self.assertTrue(hasattr(self.metadata_mapper, "mappings"))

    def test_rename_key_via_mapping(self):
        self.metadata_mapper.metadata["old"] = "foo"
        mapping = [["", "rename_key", ["old", "new"]]]
        self.metadata_mapper.mappings = mapping
        self.metadata_mapper.map()
        self.assertTrue("new" in self.metadata_mapper.metadata.keys())
        self.assertFalse("old" in self.metadata_mapper.metadata.keys())

    def test_rename_key_on_second_level_via_mapping(self):
        self.metadata_mapper.metadata["test"] = {"old": "foo"}
        mapping = [["test", "rename_key", ["old", "new"]]]
        self.metadata_mapper.mappings = mapping
        self.metadata_mapper.map()
        self.assertTrue("new" in self.metadata_mapper.metadata["test"].keys())
        self.assertFalse(
            "old" in self.metadata_mapper.metadata["test"].keys()
        )

    def test_rename_key_with_not_existing_dict_ignores(self):
        self.metadata_mapper.metadata["old"] = "foo"
        mapping = [["test", "rename_key", ["old", "new"]]]
        self.metadata_mapper.mappings = mapping
        self.metadata_mapper.map()

    def test_rename_key_with_not_existing_dict_logs(self):
        self.metadata_mapper.metadata["old"] = "foo"
        mapping = [["test", "rename_key", ["old", "new"]]]
        self.metadata_mapper.mappings = mapping
        with self.assertLogs(__package__, level="DEBUG") as cm:
            self.metadata_mapper.map()
        self.assertIn('Key "{}" not found'.format("test"), cm.output[0])

    def test_combine_items_via_mapping(self):
        old_keys = {"key1": "bla", "key2": "blub"}
        new_key = "new_key"
        mapping = [["", "combine_items", [["key1", "key2"], "new_key"]]]
        self.metadata_mapper.mappings = mapping
        for key in old_keys.keys():
            self.metadata_mapper.metadata[key] = old_keys[key]
        self.metadata_mapper.map()

        self.assertTrue(new_key in self.metadata_mapper.metadata.keys())
        for key in old_keys.keys():
            self.assertFalse(key in self.metadata_mapper.metadata.keys())
        self.assertEqual(self.metadata_mapper.metadata[new_key], "blablub")

    def test_combine_items_joint_by_pattern_via_mapping(self):
        old_keys = {"key1": "bla", "key2": "blub"}
        new_key = "new_key"
        mapping = [["", "combine_items", [["key1", "key2"], "new_key", " "]]]
        self.metadata_mapper.mappings = mapping
        for key in old_keys.keys():
            self.metadata_mapper.metadata[key] = old_keys[key]
        self.metadata_mapper.map()

        self.assertTrue(new_key in self.metadata_mapper.metadata.keys())
        for key in old_keys.keys():
            self.assertFalse(key in self.metadata_mapper.metadata.keys())
        self.assertEqual(self.metadata_mapper.metadata[new_key], "bla blub")

    def test_combine_items_on_second_level_via_mapping(self):
        old_keys = {"key1": "bla", "key2": "blub"}
        new_key = "new_key"
        mapping = [["test", "combine_items", [["key1", "key2"], "new_key"]]]
        self.metadata_mapper.mappings = mapping
        self.metadata_mapper.metadata["test"] = dict()
        for key in old_keys.keys():
            self.metadata_mapper.metadata["test"][key] = old_keys[key]
        self.metadata_mapper.map()

        self.assertTrue(
            new_key in self.metadata_mapper.metadata["test"].keys()
        )
        for key in old_keys.keys():
            self.assertFalse(
                key in self.metadata_mapper.metadata["test"].keys()
            )
        self.assertEqual(
            self.metadata_mapper.metadata["test"][new_key], "blablub"
        )

    def test_has_method_copy_key(self):
        self.assertTrue(callable(self.metadata_mapper.copy_key))

    def test_copy_key(self):
        new_key = "foo"
        old_key = "bar"
        self.metadata_mapper.metadata[old_key] = "bar"
        self.metadata_mapper.copy_key(old_key, new_key)
        self.assertTrue(new_key in self.metadata_mapper.metadata.keys())
        self.assertTrue(old_key in self.metadata_mapper.metadata.keys())

    def test_copy_key_via_mapping(self):
        self.metadata_mapper.metadata["old"] = "foo"
        mapping = [["", "copy_key", ["old", "new"]]]
        self.metadata_mapper.mappings = mapping
        self.metadata_mapper.map()
        self.assertTrue("new" in self.metadata_mapper.metadata.keys())
        self.assertTrue("old" in self.metadata_mapper.metadata.keys())

    def test_copy_key_on_second_level_via_mapping(self):
        self.metadata_mapper.metadata["test"] = {"old": "foo"}
        mapping = [["test", "copy_key", ["old", "new"]]]
        self.metadata_mapper.mappings = mapping
        self.metadata_mapper.map()
        self.assertTrue("new" in self.metadata_mapper.metadata["test"].keys())
        self.assertTrue("old" in self.metadata_mapper.metadata["test"].keys())

    def test_has_method_move_item(self):
        self.assertTrue(callable(self.metadata_mapper.move_item))

    def test_move_item(self):
        key = "foobar"
        source = "foo"
        target = "bar"
        self.metadata_mapper.metadata = {"foo": {"foobar": "baz"}, "bar": {}}
        self.metadata_mapper.move_item(key, source, target)
        self.assertTrue(key in self.metadata_mapper.metadata[target].keys())
        self.assertFalse(key in self.metadata_mapper.metadata[source].keys())

    def test_move_item_without_source(self):
        key = "foobar"
        source = ""
        target = "bar"
        self.metadata_mapper.metadata = {"foobar": "baz", "bar": {}}
        self.metadata_mapper.move_item(key, source, target)
        self.assertTrue(key in self.metadata_mapper.metadata[target].keys())
        self.assertFalse(key in self.metadata_mapper.metadata.keys())

    def test_move_item_with_not_existing_key_ignores(self):
        key = "foobar"
        source = ""
        target = "bar"
        self.metadata_mapper.metadata = {"foo": "baz", "bar": {}}
        self.metadata_mapper.move_item(key, source, target)

    def test_move_item_with_not_existing_key_logs(self):
        key = "foobar"
        source = ""
        target = "bar"
        self.metadata_mapper.metadata = {"foo": "baz", "bar": {}}
        with self.assertLogs(__package__, level="DEBUG") as cm:
            self.metadata_mapper.move_item(key, source, target)
        self.assertIn(
            'Key "{}" not found in "{}"'.format(key, source), cm.output[0]
        )

    def test_move_item_via_mapping(self):
        key = "foobar"
        source = "foo"
        target = "bar"
        self.metadata_mapper.metadata = {"foo": {"foobar": "baz"}, "bar": {}}
        mapping = [["", "move_item", [key, source, target]]]
        self.metadata_mapper.mappings = mapping
        self.metadata_mapper.map()
        self.assertTrue(key in self.metadata_mapper.metadata[target].keys())
        self.assertFalse(key in self.metadata_mapper.metadata[source].keys())

    def test_move_item_on_second_level_via_mapping(self):
        key = "foobar"
        source = "foo"
        target = "bar"
        self.metadata_mapper.metadata["test"] = {
            "foo": {"foobar": "baz"},
            "bar": {},
        }
        mapping = [["test", "move_item", [key, source, target]]]
        self.metadata_mapper.mappings = mapping
        self.metadata_mapper.map()
        self.assertTrue(
            key in self.metadata_mapper.metadata["test"][target].keys()
        )
        self.assertFalse(
            key in self.metadata_mapper.metadata["test"][source].keys()
        )

    def test_move_item_via_mapping_creating_target_dict(self):
        key = "foobar"
        source = "foo"
        target = "bar"
        self.metadata_mapper.metadata = {"foo": {"foobar": "baz"}}
        mapping = [["", "move_item", [key, source, target, True]]]
        self.metadata_mapper.mappings = mapping
        self.metadata_mapper.map()
        self.assertTrue(key in self.metadata_mapper.metadata[target].keys())
        self.assertFalse(key in self.metadata_mapper.metadata[source].keys())

    def test_move_item_via_mapping_to_nonexistent_dict_raises(self):
        key = "foobar"
        source = "foo"
        target = "bar"
        self.metadata_mapper.metadata = {"foo": {"foobar": "baz"}}
        mapping = [["", "move_item", [key, source, target, False]]]
        self.metadata_mapper.mappings = mapping
        with self.assertRaises(KeyError):
            self.metadata_mapper.map()

    def test_has_method_create_mappings(self):
        self.assertTrue(callable(self.metadata_mapper.create_mappings))

    def test_create_mappings_without_recipe_filename_raises(self):
        with self.assertRaises(exceptions.MissingFilenameError):
            self.metadata_mapper.create_mappings()

    def test_create_mappings_with_missing_version_raises(self):
        self.metadata_mapper.recipe_filename = self.filename
        self.metadata_mapper.version = "0.1.0"
        self.write_metadata_mapper_file(dict())
        with self.assertRaises(exceptions.MissingRecipeError):
            self.metadata_mapper.create_mappings()

    def test_create_mappings_with_existing_version(self):
        mapping_dict = {"mapping1": {"metadata file versions": ["0.1.0"]}}
        self.metadata_mapper.recipe_filename = self.filename
        self.metadata_mapper.version = "0.1.0"
        self.write_metadata_mapper_file(mapping_dict)
        self.metadata_mapper.create_mappings()

    def test_create_mappings_with_existing_version_and_format(self):
        mapping_dict = {
            "format": {"type": "metadata mapper", "version": "0.1.0"},
            "mapping1": {"metadata file versions": ["0.1.0"]},
        }
        self.metadata_mapper.recipe_filename = self.filename
        self.metadata_mapper.version = "0.1.0"
        self.write_metadata_mapper_file(mapping_dict)
        self.metadata_mapper.create_mappings()

    def test_create_mappings_adds_combine_items_to_mappings(self):
        combine_items = [
            {
                "old keys": ["Date start", "Time start"],
                "new key": "start",
                "pattern": " ",
                "in dict": "GENERAL",
            }
        ]
        mapping = [
            "GENERAL",
            "combine_items",
            [["Date start", "Time start"], "start", " "],
        ]
        mapping_dict = {
            "format": {"type": "metadata mapper", "version": "0.1.0"},
            "mapping1": {
                "metadata file versions": ["0.1.0"],
                "combine items": combine_items,
            },
        }
        self.metadata_mapper.recipe_filename = self.filename
        self.metadata_mapper.version = "0.1.0"
        self.write_metadata_mapper_file(mapping_dict)
        self.metadata_mapper.create_mappings()
        self.assertEqual(mapping, self.metadata_mapper.mappings[-1])

    def test_create_mappings_adds_rename_key_to_mappings(self):
        rename_key = [
            {"old key": "GENERAL", "new key": "measurement", "in dict": ""}
        ]
        mapping = ["", "rename_key", ["GENERAL", "measurement"]]
        mapping_dict = {
            "format": {"type": "metadata mapper", "version": "0.1.0"},
            "mapping1": {
                "metadata file versions": ["0.1.0"],
                "rename key": rename_key,
            },
        }
        self.metadata_mapper.recipe_filename = self.filename
        self.metadata_mapper.version = "0.1.0"
        self.write_metadata_mapper_file(mapping_dict)
        self.metadata_mapper.create_mappings()
        self.assertEqual(mapping, self.metadata_mapper.mappings[-1])

    def test_create_mappings_adds_copy_key_to_mappings(self):
        copy_key = [
            {"old key": "Date", "new key": "Date end", "in dict": "GENERAL"}
        ]
        mapping = ["GENERAL", "copy_key", ["Date", "Date end"]]
        mapping_dict = {
            "format": {"type": "metadata mapper", "version": "0.1.0"},
            "mapping1": {
                "metadata file versions": ["0.1.0"],
                "copy key": copy_key,
            },
        }
        self.metadata_mapper.recipe_filename = self.filename
        self.metadata_mapper.version = "0.1.0"
        self.write_metadata_mapper_file(mapping_dict)
        self.metadata_mapper.create_mappings()
        self.assertEqual(mapping, self.metadata_mapper.mappings[-1])

    def test_create_mappings_adds_move_item_to_mappings(self):
        move_item = [
            {
                "key": "model",
                "source dict": "measurement",
                "target dict": "spectrometer",
            }
        ]
        mapping = ["", "move_item", ["model", "measurement", "spectrometer"]]
        mapping_dict = {
            "format": {"type": "metadata mapper", "version": "0.1.0"},
            "mapping1": {
                "metadata file versions": ["0.1.0"],
                "move item": move_item,
            },
        }
        self.metadata_mapper.recipe_filename = self.filename
        self.metadata_mapper.version = "0.1.0"
        self.write_metadata_mapper_file(mapping_dict)
        self.metadata_mapper.create_mappings()
        self.assertEqual(mapping, self.metadata_mapper.mappings[-1])

    def test_create_mappings_adds_move_item_with_create_to_mappings(self):
        move_item = [
            {
                "key": "model",
                "source dict": "measurement",
                "target dict": "spectrometer",
                "create target": "True",
            }
        ]
        mapping = [
            "",
            "move_item",
            ["model", "measurement", "spectrometer", True],
        ]
        mapping_dict = {
            "format": {"type": "metadata mapper", "version": "0.1.0"},
            "mapping1": {
                "metadata file versions": ["0.1.0"],
                "move item": move_item,
            },
        }
        self.metadata_mapper.recipe_filename = self.filename
        self.metadata_mapper.version = "0.1.0"
        self.write_metadata_mapper_file(mapping_dict)
        self.metadata_mapper.create_mappings()
        self.assertEqual(mapping, self.metadata_mapper.mappings[-1])

    def test_map_adds_items_to_mappings_if_filename_is_given(self):
        combine_items = [
            {
                "old keys": ["Date start", "Time start"],
                "new key": "start",
                "pattern": " ",
                "in dict": "GENERAL",
            }
        ]
        mapping = [
            "GENERAL",
            "combine_items",
            [["Date start", "Time start"], "start", " "],
        ]
        mapping_dict = {
            "format": {"type": "metadata mapper", "version": "0.1.0"},
            "mapping1": {
                "metadata file versions": ["0.1.0"],
                "combine items": combine_items,
            },
        }
        self.metadata_mapper.recipe_filename = self.filename
        self.metadata_mapper.version = "0.1.0"
        self.write_metadata_mapper_file(mapping_dict)
        self.metadata_mapper.metadata = {
            "GENERAL": {"Date start": "", "Time start": ""}
        }
        self.metadata_mapper.map()
        self.assertEqual(mapping, self.metadata_mapper.mappings[-1])



"""Input and output (IO) of information from and to the persistence layer.

Currently, input and output of both, datasets and recipes can be handled.


Datasets
========

Both, data and metadata contained in datasets as well as the information
stored in recipes for recipe-driven data analysis can be read and written.

For datasets, two generic classes are provided:

* :class:`aspecd.io.DatasetImporter`

* :class:`aspecd.io.DatasetExporter`

As the name says, these classes should be used to implement import and
export functionality for your own purposes in applications derived from the
ASpecD framework.

Generally, both import and export should be handled via the respective
methods of the :class:`aspecd.dataset.Dataset` class, thus first
instantiating an object of that class and an appropriate importer or
exporter, and afterwards only operating on the dataset using its methods.

In its most generic form, this may look something like:

.. code-block::

    dataset = aspecd.dataset.Dataset()
    importer = aspecd.io.DatasetImporter(source="/path/to/your/data")
    dataset.import_from(importer)

Similarly, you would handle the export of your data (and metadata)
contained in a dataset object using an exporter object, respectively.

.. code-block::

    dataset = aspecd.dataset.Dataset()
    importer = aspecd.io.DatasetExporter(target="/path/to/destination")
    dataset.export_to(exporter)


However, if you use :ref:`recipe-driven data analysis <recipes>`, things
become much simpler:

* Imports will be automatically taken care of.

* Exports can be specified as simple task.


A simple example of a recipe only loading datasets and afterwards exporting
them could look like this:

.. code-block:: yaml

    datasets:
      - /path/to/first/dataset
      - /path/to/second/dataset

    tasks:
      - kind: export
        type: AdfExporter
        properties:
          target:
            - dataset1
            - dataset2


What is happening here? Two datasets are imported, and afterwards exported
to the ASpecD Dataset Format (ADF) using the :class:`aspecd.io.AdfExporter`.

Another frequent use case, although one that admittedly pretty much opposes
the whole idea of the ASpecD framework in terms of reproducibility and
traceability: Your collaboration partners require you to provide them with
raw data they can import into their favourite program for creating plots.
The only viable way: export to plain text (ouch!) - saying good-bye to all
your metadata and history:

.. code-block:: yaml

    datasets:
      - /path/to/first/cool/dataset
      - /path/to/second/cool/dataset
      - /path/to/another/cool/dataset

    tasks:
      - kind: export
        type: TxtExporter
        properties:
          target:
            - cool-dataset1
            - cool-dataset2
            - cool-dataset3


In this case, you can as well add whatever processing necessary to your
datasets before exporting them, and you see that recipes come in quite handy
here.


More control over imports
-------------------------

Sometimes there is the need to have more control over the import,
be it that you would want to set labels for datasets explicitly upon load,
determine which importer to use, or provide additional parameters for an
importer (a frequent use case for the rather generic :class:`TxtImporter`).

This is an excerpt of an example recipe importing ASCII exports of a
common UV/Vis spectrometer and showing many of the options possible:

.. code-block:: yaml

    datasets:
      - source: cbztbt.txt
        label: D-A
        id: Cbz-TBT
        importer: TxtImporter
        importer_parameters:
          skiprows: 2
          separator: ','

So what's happening here? Lets go through step by step:

* Datasets are a list, as usual, but this time, it is not a list of
  filenames, but a list of (hierarchical) key--value pairs.
* The ``source`` key sets the filename (and can include a path, as usual).
* The ``label`` key sets the label used for the dataset, *e.g.*,
  in a figure legend.
* The ``id`` key sets the (unique) identifier (ID) the dataset can be
  referred to throughout the recipe. This is often useful if you want to
  restrict certain tasks to only a subset of the loaded datasets.
* The ``importer`` key sets the importer class to use. This class needs to
  be available from within your current package. You can prefix the class
  name with a package if you like.
* The ``importer_parameters`` key is a series of key--value pairs (*i.e.*,
  a :class:`dict` in Python language) setting additional parameters for the
  specific importer. See the documentation of the respective importer class
  for further details.

Of course, you need not use all of these parameters. Usually, if you want
to specify importer parameters, it is a good idea to be explicit about the
importer as well. However, even that is not strictly necessary. The only
thing that is strictly necessary: As soon as you want to provide more than
a filename/path per dataset, you need to switch from a list of strings (
*i.e.*, filenames/paths) to a key--value approach, with ``source`` being
the key for the filename/path.


Importers for specific file formats
-----------------------------------

There exists a series of importers for specific file formats:

* :class:`aspecd.io.AdfImporter`

  Importer for data in ASpecD Dataset Format (ADF)

* :class:`aspecd.io.AsdfImporter`

  Importer for data in asdf format

* :class:`aspecd.io.TxtImporter`

  Importer for data in plain text format

For details, see the respective class documentation.


Exporters for specific file formats
-----------------------------------

Datasets need to be persisted sometimes, and currently, there exist two
exporters for specific file formats that can be imported again using the
respective importers. Furthermore, the full information contained in a
dataset will be retained.

* :class:`aspecd.io.AdfExporter`

  Exporter for datasets to ASpecD Dataset Format (ADF)

* :class:`aspecd.io.AsdfExporter`

  Exporter for datasets to asdf format

For details, see the respective class documentation.

A bit a special case is the exporter to plain text files, as this file
format does *not* preserve the metadata stored within the dataset and should
only be used as last resort:

* :class:`aspecd.io.TxtExporter`

  Exporter for data to plain text format


.. warning::
    All metadata contained within a dataset (including the full history)
    are lost when exporting to plain text. Therefore, using this
    exporter will usually result in you loosing reproducibility. Hence,
    better think twice before using this exporter and use entirely on
    your own risk and only if you *really* know what you are doing (and
    why).


Writing importers for data
--------------------------

When writing importer classes for your own data, there is a number of
pitfalls, some of which shall be described here together with solutions and
"best practices".

Dimensions of data
~~~~~~~~~~~~~~~~~~

Usually, we assign axes in the order *x*, *y*, *z*, and assume the *x* axis
to be the horizontal axis in a plot. However, numpy (as well as other
software), follows a different convention, with the first index referring to
the *row* of your matrix, the second index to the *column*. That boils down
to having the first index correspond to the *y* axis, and the second index
referring to the *x* axis.

As long as your data are one-dimensional, resulting in two axes objects in
your dataset, everything is fine, and the second axis will have no values.

However, if your data to be imported are two-dimensional, your first
dimension will be the index of rows (along a column), hence the *y* axis,
and the second dimension the index of your columns (along a row), *i.e.* the
*x* axis. This is perfectly fine, and it is equally fine to revert this
order, as long as you ensure your axis objects to be consistent with the
dimensions of your data.

If you assign numeric data to the :attr:`aspecd.dataset.Data.data` property,
the corresponding axes values will initially be set to the indices of the
data points along the corresponding dimension, with the first axis (index 0)
corresponding to the first dimension (row indices along a column) and
similar for each of the following dimensions of your data. Note that there
will always be one axis more than dimensions of your data. This last axis
will not have values, and usually its quantity is something like "intensity".


Backup of the data
~~~~~~~~~~~~~~~~~~

One essential concept of the ASpecD dataset is to store the original data
together with their axes in a separate, non-public property. This is done
automatically by the importer after calling out to its non-public method
:meth:`aspecd.io.DatasetImporter._import`. Hence, usually you need not take
care of this at all.


Handling of metadata
~~~~~~~~~~~~~~~~~~~~

Data without information about these data are usually pretty useless. Hence,
an ASpecD dataset is always a unit of numerical data and corresponding
metadata. While you will need to come up with your own structure for
metadata of your datasets and create a hierarchy of classes derived from
:class:`aspecd.metadata.DatasetMetadata`, your importers need to ensure that
these metadata are populated respectively. Of course, which metadata can be
populated depends strongly on the file format you are about to import.


Handling different file formats for importing data
--------------------------------------------------

Often, data are available in different formats, and deciding which importer
is appropriate for a given format can be quite involved. To free other
classes from having to contain the relevant code, a factory can be used:

* :class:`aspecd.io.DatasetImporterFactory`

Currently, the sole information provided to decide about the appropriate
importer is the source (a string). A concrete importer object is returned
by the method :meth:`get_importer`. Thus, using the factory in another
class may look like the following::

  importer_factory = aspecd.io.DatasetImporterFactory()
  importer = importer_factory.get_importer(source="/path/to/your/data")
  dataset = aspecd.dataset.Dataset()
  dataset.import_from(importer)

Here, as in the example above, "source" refers to a (unique) identifier of
your dataset, be it a filename, path, URL/URI, LOI, or alike.

.. important::
    For recipe-driven data analysis to work with an ASpecD-derived package,
    you need to implement a :class:`aspecd.io.DatasetImporterFactory` class
    there as well that can be obtained by instantiating
    ``<your_package>.io.DatasetImporterFactory()``.


Recipes
=======

For recipes, a similar set of classes is provided:

* :class:`aspecd.io.RecipeImporter`

* :class:`aspecd.io.RecipeExporter`

For additional concrete classes handling import and export from and to YAML
files see below.

The same general principles laid out above for the datasets applies to
these classes as well. In particular, both import and export should be
handled via the respective methods of the :class:`aspecd.tasks.Recipe`
class, thus first instantiating an object of that class and an appropriate
importer or exporter, and afterwards only operating on the recipe using
its methods.

In its most generic form, this may look something like::

    recipe = aspecd.tasks.Recipe()
    importer = aspecd.io.RecipeImporter(source="/path/to/your/recipe")
    recipe.import_from(importer)

Similarly, you would handle the export of the information contained in a
recipe object using an exporter object, respectively.

To simplify the input and output of recipes, and due recipe-driven data
analysis being an intrinsic property of the ASpecD framework, two classes
handling the import and export from and to YAML files are provided as well:

* :class:`aspecd.io.RecipeYamlImporter`

* :class:`aspecd.io.RecipeYamlExporter`

These classes can directly be used to work with YAML files containing
information for recipe-driven data analysis. For details of the YAML file
structure, see the :class:`aspecd.tasks.Recipe` class and its attributes.


Module documentation
====================

"""

import copy
import io
import logging
import os
import tempfile
import zipfile

import asdf
import numpy as np

import aspecd.exceptions
import aspecd.history
import aspecd.metadata
import aspecd.utils

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class DatasetImporter:
    """Base class for dataset importer.

    Each class actually importing data and metadata into a dataset should
    inherit from this class.

    To perform the import, call the
    :meth:`~aspecd.dataset.Dataset.import_from` method of the dataset
    the import should be performed for, and provide a reference to the
    actual importer object to it.

    The actual implementation of the importing is done in the private method
    :meth:`_import` that in turn gets called by :meth:`import_into`
    which is called by the :meth:`aspecd.dataset.Dataset.import_from` method
    of the dataset object.

    One question arising when actually implementing an importer for a
    specific file format: How do the data get into the dataset? The simple
    answer: The :meth:`_import` method of the importer knows about the
    dataset and its structure (see :class:`aspecd.dataset.Dataset` for
    details) and assigns data (and metadata) read from an external source
    to the respective fields of the dataset. In terms of a broader software
    architecture point of view: The dataset knows nothing about the
    importer besides its bare existence and interface, whereas the importer
    knows about the dataset and how to map data and metadata.

    Attributes
    ----------
    dataset : :class:`aspecd.dataset.Dataset`
        dataset to import data and metadata into

    source : :class:`str`
        specifier of the source the data and metadata will be read from

    parameters : :class:`dict`
        Additional parameters to control import options.

        Useful in case of, *e.g.*, CSV importers where the user may want to
        set things such as the delimiter

        .. versionadded:: 0.2

    Raises
    ------
    aspecd.io.MissingDatasetError
        Raised when no dataset exists to act upon

    """

    def __init__(self, source=None):
        self.source = source
        self.dataset = None
        self.parameters = {}

    def import_into(self, dataset=None):
        """Perform the actual import into the given dataset.

        If no dataset is provided at method call, but is set as property in
        the importer object, the :meth:`aspecd.dataset.Dataset.import_from`
        method of the dataset will be called.

        If no dataset is provided at method call nor as property in the
        object, the method will raise a respective exception.

        The dataset object always calls this method with the respective
        dataset as argument. Therefore, in this case setting the dataset
        property within the importer object is not necessary.

        The actual import should be implemented within the non-public method
        :meth:`_import`.

        .. note::
            A number of parameters of the dataset are automatically assigned
            *after* calling out to the non-public method
            :meth:`aspecd.io.DatasetImporter._import`, namely the
            non-public property ``_origdata`` of the dataset is populated
            with a copy of :attr:`aspecd.dataset.Dataset.data`, and id and
            label are set to :attr:`aspecd.io.DatasetImporter.source`.

        Parameters
        ----------
        dataset : :class:`aspecd.dataset.Dataset`
            Dataset to import data and metadata into

        Raises
        ------
        aspecd.io.MissingDatasetError
            Raised if no dataset is provided.

        """
        if not dataset:
            if self.dataset:
                self.dataset.import_from(self)
            else:
                raise aspecd.exceptions.MissingDatasetError(
                    "No dataset provided"
                )
        else:
            self.dataset = dataset
        self._import()
        # Untested due to lack of ideas how to test
        # pylint: disable=protected-access
        self.dataset._origdata = copy.deepcopy(self.dataset.data)
        self.dataset.id = self.source
        if self.source and not self.dataset.label:
            self.dataset.label = os.path.split(self.source)[-1]

    def _import(self):
        """Perform the actual import of data and metadata into the dataset.

        The implementation of the actual import goes in here in all
        classes inheriting from DatasetImporter. This method is automatically
        called by :meth:`import_into`.

        Importing data and metadata includes assigning both to the respective
        fields of the :obj:`aspecd.dataset.Dataset` object. For details of
        its structure, see there.

        Usually, this method will successively call other private/protected
        methods of the importer to perform the required tasks that are
        specific for each data source.

        """


class DatasetImporterFactory:
    """
    Factory for creating importer objects based on the source provided.

    Often, data are available in different formats, and deciding which
    importer is appropriate for a given format can be quite involved. To
    free other classes from having to contain the relevant code, a factory
    can be used.

    Currently, the sole information provided to decide about the
    appropriate importer is the source (a string). A concrete importer
    object is returned by the method :meth:`get_importer`. If no source is
    provided, an exception will be raised.

    The actual code for deciding which type of importer to return in what
    case should be implemented in the non-public method :meth:`_get_importer`
    in any package based on the ASpecD framework.

    In its basic implementation, as done here, the non-public method
    :meth:`_get_importer` returns the importers for ADF, ASDF, and TXT
    depending on the file extension, and in all other cases the standard
    importer.

    This might be a viable way for an own :class:`DatasetImporterFactory`
    implementation in the rare case of having only one single type of data,
    but provides a sensible starting point for own developments.


    Attributes
    ----------
    source : :class:`str`
        Source of the dataset to be loaded.

        Gets set by calling the method :meth:`get_importer` with the
        ``source`` parameter.

    Raises
    ------
    aspecd.io.MissingSourceError
        Raised if no source is provided

    """

    def __init__(self):
        self.source = None

    def get_importer(self, source="", importer="", parameters=None):
        """
        Return importer object for dataset specified by its source.

        The actual code for deciding which type of importer to return in what
        case should be implemented in the non-public method
        :meth:`_get_importer` in any package based on the ASpecD framework.

        If no importer gets returned by the method :meth:`_get_importer`,
        the ASpecD-interal importers will be checked for matching the file
        type. Thus, you can overwrite the behaviour of any filetype
        supported natively by the ASpecD framework, but retain compatibility
        to the ASpecD-specific file types.

        .. note::
            Currently, only filenames/paths are supported, and if ``source``
            does not start with the file separator, the absolute path to the
            current directory is prepended.


        Parameters
        ----------
        source : :class:`str`
            string describing the source of the dataset

            May be a filename or path, a URL/URI, a LOI, or similar

        importer : :class:`str`
            Name of the importer to use for importing the dataset

            Default: ''

            .. versionadded:: 0.2

        parameters : :class:`dict`
            Additional parameters for controlling the import

            Default: None

            .. versionadded:: 0.2

        Returns
        -------
        importer : :class:`aspecd.io.DatasetImporter`
            importer object of appropriate class

        Raises
        ------
        aspecd.io.MissingSourceError
            Raised if no source is provided

        """
        if not source:
            raise aspecd.exceptions.MissingSourceError(
                "A source is required to return an appropriate importer"
            )
        self.source = source
        if not self.source.startswith(os.pathsep):
            self.source = os.path.join(
                os.path.abspath(os.curdir), self.source
            )
        if importer:
            package_name = aspecd.utils.package_name(self)
            # Currently untested
            if not package_name.endswith("io"):
                package_name = ".".join([package_name, "io"])
            full_class_name = ".".join([package_name, importer])
            importer = aspecd.utils.object_from_class_name(full_class_name)
            importer.source = self.source
        if not importer:
            importer = self._get_importer()
        if not importer:
            importer = self._get_aspecd_importer()
        if parameters:
            importer.parameters = parameters
        return importer

    # noinspection PyMethodMayBeStatic
    def _get_importer(self):
        """Choose appropriate importer for a dataset.

        Every package inheriting from the ASpecD framework should implement
        this method. Note that in case you do not handle a filetype and
        hence return no importer, the default ASpecD importer will be
        checked for matching the given source. Thus, you can overwrite the
        behaviour of any filetype supported natively by the ASpecD
        framework, but retain compatibility to the ASpecD-specific file types.

        Returns
        -------
        importer : :class:`aspecd.io.DatasetImporter`
            Importer for the specific file type

        """
        importer = None
        return importer

    def _get_aspecd_importer(self):
        _, file_extension = os.path.splitext(self.source)
        if file_extension == ".adf":
            return AdfImporter(source=self.source)
        if file_extension == ".asdf":
            return AsdfImporter(source=self.source)
        if file_extension == ".txt":
            return TxtImporter(source=self.source)
        logger.warning(
            "No importer found. Using default importer. This may "
            "result in downstream problems."
        )
        return DatasetImporter(source=self.source)


class DatasetExporter:
    """Base class for dataset exporter.

    Each class actually exporting data from a dataset to some other should
    inherit from this class.

    To perform the export, call the
    :meth:`~aspecd.dataset.Dataset.export_to` method of the dataset
    the export should be performed for, and provide a reference to the
    actual exporter object to it.

    The actual implementation of the exporting is done in the non-public
    method :meth:`_export` that in turn gets called by :meth:`export_from`
    which is called by the :meth:`aspecd.dataset.Dataset.export_to` method
    of the dataset object.

    Attributes
    ----------
    dataset : :obj:`aspecd.dataset.Dataset`
        dataset to export data and metadata from

    target : string
        specifier of the target the data and metadata will be written to

    comment : :class:`str`
        User-supplied comment describing intent, purpose, reason, ...


    Raises
    ------
    aspecd.io.MissingDatasetError
        Raised when no dataset exists to act upon


    .. versionchanged:: 0.6.4
        New attribute :attr:`comment`

    """

    def __init__(self, target=None):
        self.target = target
        self.dataset = None
        self.comment = ""

    def export_from(self, dataset=None):
        """Perform the actual export from the given dataset.

        If no dataset is provided at method call, but is set as property in
        the exporter object, the :meth:`aspecd.dataset.Dataset.export_to`
        method of the dataset will be called.

        If no dataset is provided at method call nor as property in the
        object, the method will raise a respective exception.

        The dataset object always calls this method with the respective
        dataset as argument. Therefore, in this case setting the dataset
        property within the exporter object is not necessary.

        The actual export is implemented within the non-public method
        :meth:`_export` that gets automatically called.

        Parameters
        ----------
        dataset : :class:`aspecd.dataset.Dataset`
            Dataset to export data and metadata from

        Raises
        ------
        aspecd.io.MissingDatasetError
            Raised if no dataset is provided.

        """
        if not dataset:
            if self.dataset:
                self.dataset.export_to(self)
            else:
                raise aspecd.exceptions.MissingDatasetError(
                    "No dataset provided"
                )
        else:
            self.dataset = dataset
        self._export()

    def _export(self):
        """Perform the actual export of data and metadata from the dataset.

        The implementation of the actual export goes in here in all
        classes inheriting from DatasetExporter. This method is automatically
        called by :meth:`export_from`.

        Usually, this method will successively call other private/protected
        methods of the exporter to perform the required tasks that are
        specific for each target format.

        """

    def create_history_record(self):
        """
        Create history record to be added to the dataset.

        Usually, this method gets called from within the
        :meth:`aspecd.dataset.export_to` method of the
        :class:`aspecd.dataset.Dataset` class and ensures the history of
        each processing step to get written properly.

        Returns
        -------
        history_record : :class:`aspecd.history.DatasetExporterHistoryRecord`
            history record for export step


        .. versionadded:: 0.9

        """
        history_record = aspecd.history.DatasetExporterHistoryRecord(
            package=self.dataset.package_name, exporter=self
        )
        return history_record


class RecipeImporter:
    """Base class for recipe importer.

    Each class actually importing recipes into a :obj:`aspecd.tasks.Recipe`
    object should inherit from this class.

    To perform the import, call the
    :meth:`~aspecd.tasks.Recipe.import_from` method of the recipe the
    import should be performed for, and provide a reference to the
    actual importer object to it.

    The actual implementation of the importing is done in the non-public
    method :meth:`_import` that in turn gets called by :meth:`import_into`
    which is called by the :meth:`aspecd.tasks.Recipe.import_from` method
    of the recipe object.

    One question arising when actually implementing an importer for a
    specific file format: How does the information get into the recipe? The
    simple answer: The :meth:`_import` method of the importer knows about the
    recipe and its structure (see :class:`aspecd.tasks.Recipe` for
    details) and creates a dictionary with keys corresponding to the
    respective attributes of the recipe. In turn, it can then call the
    :meth:`aspecd.tasks.Recipe.from_dict` method. In terms of a broader
    software architecture point of view: The recipe knows nothing about the
    importer besides its bare existence and interface, whereas the importer
    knows about the recipe and how to map the information obtained to it.

    Attributes
    ----------
    recipe : :obj:`aspecd.tasks.Recipe`
        recipe to import into
    source : :class:`str`
        specifier of the source the information will be read from

    Raises
    ------
    aspecd.io.MissingRecipeError
        Raised when no dataset exists to act upon

    """

    def __init__(self, source=""):
        self.source = source
        self.recipe = None

    def import_into(self, recipe=None):
        """Perform the actual import into the given recipe.

        If no recipe is provided at method call, but is set as property in
        the importer object, the :meth:`aspecd.tasks.Recipe.import_from`
        method of the recipe will be called.

        If no recipe is provided at method call nor as property in the
        object, the method will raise a respective exception.

        The recipe object always calls this method with the respective
        recipe as argument. Therefore, in this case setting the recipe
        property within the importer object is not necessary.

        The actual import should be implemented within the non-public method
        :meth:`_import`.

        Parameters
        ----------
        recipe : :obj:`aspecd.tasks.Recipe`
            recipe to import into

        Raises
        ------
        aspecd.io.MissingRecipeError
            Raised if no recipe is provided.

        """
        if not recipe:
            if self.recipe:
                self.recipe.import_from(self)
            else:
                raise aspecd.exceptions.MissingRecipeError(
                    "No recipe provided"
                )
        else:
            self.recipe = recipe
        self.recipe.filename = self.source
        self._import()

    def _import(self):
        """Perform the actual import into the recipe.

        The implementation of the actual import goes in here in all
        classes inheriting from RecipeImporter. This method is automatically
        called by :meth:`import_into`.

        Importing metadata includes assigning it to the respective fields
        of the :obj:`aspecd.tasks.Recipe` object. For details of
        its structure, see there. To do this, the method should create a
        dictionary that can afterwards be supplied as an argument to a call
        to :meth:`aspecd.tasks.Recipe.from_dict`.

        """


class RecipeExporter:
    """Base class for recipe exporter.

    Each class actually exporting recipes from :obj:`aspecd.tasks.Recipe`
    objects should inherit from this class.

    To perform the export, call the
    :meth:`aspecd.tasks.Recipe.export_to` method of the recipe the export
    should be performed for, and provide a reference to the actual exporter
    object to it.

    The actual implementation of the exporting is done in the non-public
    method :meth:`_export` that in turn gets called by :meth:`export_from`
    which is called by the :meth:`aspecd.tasks.Recipe.export_to` method
    of the recipe object.

    Attributes
    ----------
    recipe : :obj:`aspecd.tasks.Recipe`
        recipe to export information from
    target : string
        specifier of the target the information will be written to

    Raises
    ------
    aspecd.io.MissingRecipeError
        Raised when no dataset exists to act upon

    """

    def __init__(self, target=""):
        self.target = target
        self.recipe = None

    def export_from(self, recipe=None):
        """Perform the actual export from the given recipe.

        If no recipe is provided at method call, but is set as property in
        the exporter object, the :meth:`aspecd.tasks.Recipe.export_to`
        method of the recipe will be called.

        If no recipe is provided at method call nor as property in the
        object, the method will raise a respective exception.

        The recipe object always calls this method with the respective
        recipe as argument. Therefore, in this case setting the recipe
        property within the exporter object is not necessary.

        The actual export should be implemented within the non-public method
        :meth:`_export`.

        Parameters
        ----------
        recipe : :class:`aspecd.tasks.Recipe`
            Recipe to export from

        Raises
        ------
        aspecd.io.MissingRecipeError
            Raised if no recipe is provided.

        """
        if not recipe:
            if self.recipe:
                self.recipe.export_to(self)
            else:
                raise aspecd.exceptions.MissingRecipeError(
                    "No recipe provided"
                )
        else:
            self.recipe = recipe
        self._export()

    def _export(self):
        """Perform the actual export from the recipe.

        The implementation of the actual export goes in here in all
        classes inheriting from RecipeExporter. This method is automatically
        called by :meth:`export_from`.

        Usually, this method will first create a dictionary from the recipe
        using the :meth:`aspecd.tasks.Recipe.to_dict` method. This
        dictionary can afterwards be further processed and written to some
        file.

        """


class RecipeYamlImporter(RecipeImporter):
    """
    Recipe importer for importing from YAML files.

    The YAML file needs to have a structure compatible to the actual
    recipe, such that the dict created from reading the YAML file can be
    directly fed into the :meth:`aspecd.tasks.Recipe.from_dict` method.

    The order of entries of the YAML file is preserved due to using ordered
    dictionaries (:class:`collections.OrderedDict`) internally.

    Parameters
    ----------
    source : :class:`str`
        filename of a YAML file to read from

    """

    def __init__(self, source=""):
        self.recipe_version = ""
        self._recipe_dict = None
        super().__init__(source=source)

    def _import(self):
        self._load_from_yaml()
        self._convert()
        self.recipe.from_dict(self._recipe_dict)

    def _load_from_yaml(self):
        yaml = aspecd.utils.Yaml()
        yaml.read_from(filename=self.source)
        yaml.deserialise_numpy_arrays()
        self._recipe_dict = yaml.dict

    def _convert(self):
        self._get_recipe_version()
        self._map_recipe_structure()

    def _get_recipe_version(self):
        self.recipe_version = self.recipe.format["version"]
        if (
            "format" in self._recipe_dict
            and "version" in self._recipe_dict["format"]
        ):
            self.recipe_version = self._recipe_dict["format"]["version"]
        deprecated_keys = [
            "default_package",
            "autosave_plots",
            "output_directory",
            "datasets_source_directory",
        ]
        if any(key in self._recipe_dict for key in deprecated_keys):
            self.recipe_version = "0.1"

    def _map_recipe_structure(self):
        mapper = aspecd.metadata.MetadataMapper()
        mapper.version = self.recipe_version
        mapper.metadata = self._recipe_dict
        mapper.recipe_filename = "recipe_mapper.yaml"
        mapper.map()
        self._recipe_dict = mapper.metadata


class RecipeYamlExporter(RecipeExporter):
    """
    Recipe exporter for exporting to YAML files.

    The YAML file will have a structure corresponding to the output of the
    :meth:`aspecd.tasks.Recipe.to_dict` method of the recipe object.

    Parameters
    ----------
    target : :class:`str`
        filename of a YAML file to write to

    """

    def __init__(self, target=""):
        super().__init__(target=target)

    def _export(self):
        yaml = aspecd.utils.Yaml()
        yaml.dict = self.recipe.to_dict()
        yaml.numpy_array_to_list = True
        yaml.serialise_numpy_arrays()
        yaml.write_to(filename=self.target)


class AdfExporter(DatasetExporter):
    """
    Dataset exporter for exporting to ASpecD dataset format.

    The ASpecD dataset format is vaguely reminiscent of the Open Document
    Format, *i.e.* a zipped directory containing structured data (in this
    case in form of a YAML file) and binary data in a corresponding
    subdirectory.

    As PyYAML is not capable of dealing with NumPy arrays out of the box,
    those are dealt with separately. Small arrays are stored inline as
    lists, larger arrays in separate files. For details, see the
    :class:`aspecd.utils.Yaml` class.

    The data format tries to be as self-contained as possible,
    using standard file formats and a brief description of its layout
    contained within the archive. Collecting the contents in a single ZIP
    archive allows the user to deal with a single file for a dataset,
    while more advanced users can easily dig into the details and write
    importers for other platforms and programming languages, making the
    format rather platform-independent and future-safe. Due to using binary
    representation for larger numerical arrays, the format should be more
    memory-efficient than other formats.

    """

    def __init__(self, target=None):
        super().__init__(target=target)
        self.extension = ".adf"
        self._filenames = {
            "dataset": "dataset.yaml",
            "version": "VERSION",
            "readme": "README",
        }
        self._bin_dir = "binaryData"
        self._tempdir_name = ""
        self._version = "1.0.0"

    def _export(self):
        if not self.target:
            raise aspecd.exceptions.MissingTargetError
        with tempfile.TemporaryDirectory() as tempdir:
            self._tempdir_name = tempdir
            self._create_files()
            self._create_zip_archive()

    def _create_zip_archive(self):
        with zipfile.ZipFile(
            self.target + self.extension, "w"
        ) as zipped_file:
            for filename in self._filenames.values():
                zipped_file.write(
                    filename=os.path.join(self._tempdir_name, filename),
                    arcname=filename,
                )
            bin_dir_path = os.path.join(self._tempdir_name, self._bin_dir)
            zipped_file.write(
                filename=os.path.join(bin_dir_path), arcname=self._bin_dir
            )
            for filename in os.listdir(bin_dir_path):
                zipped_file.write(
                    filename=os.path.join(bin_dir_path, filename),
                    arcname=os.path.join(self._bin_dir, filename),
                )

    def _create_files(self):
        self._create_dataset_yaml()
        self._create_version_file()
        self._create_readme_file()

    def _create_dataset_yaml(self):
        bin_dir_path = os.path.join(self._tempdir_name, self._bin_dir)
        os.mkdir(bin_dir_path)
        yaml = aspecd.utils.Yaml()
        yaml.binary_directory = bin_dir_path
        yaml.dict = self.dataset.to_dict()
        yaml.serialise_numpy_arrays()
        yaml.write_to(
            filename=os.path.join(
                self._tempdir_name, self._filenames["dataset"]
            )
        )

    def _create_version_file(self):
        with open(
            os.path.join(self._tempdir_name, self._filenames["version"]),
            "w+",
            encoding="utf8",
        ) as file:
            file.write(self._version)

    def _create_readme_file(self):
        readme_contents = (
            "Readme\n"
            "======\n\n"
            "This directory contains an ASpecD dataset stored in the\n"
            "ASpecD dataset format (adf).\n\n"
            "What follows is a bit of information on the meaning of\n"
            "each of the files in the directory.\n"
            "Sources of further information on the file format\n"
            "are provided at the end of the file.\n\n"
            "Copyright (c) 2021, Till Biskup\n"
            "2021-01-04\n\n"
            "Files and their meaning\n"
            "-----------------------\n\n"
            "* dataset.yaml - text/YAML\n"
            "  hierarchical metadata store\n\n"
            "* binaryData/<filename>.npy - NumPy binary\n"
            "  numerical data of the dataset stored in NumPy format\n\n"
            "  Only arrays exceeding a certain threshold are stored\n"
            "  in binary format, mainly to save space and preserve\n"
            "  numerical accuracy.\n\n"
            "* VERSION - text\n"
            "  version number of the dataset format\n\n"
            "  The version number follows the semantic versioning scheme.\n\n"
            "* README - text\n"
            "  This file\n\n"
            "Further information\n"
            "-------------------\n\n"
            "More information can be found on the web in the\n"
            "ASpecD package documentation:\n\n"
            "https://docs.aspecd.de/adf.html\n"
        )
        with open(
            os.path.join(self._tempdir_name, self._filenames["readme"]),
            "w+",
            encoding="utf8",
        ) as file:
            file.write(readme_contents)


class AdfImporter(DatasetImporter):
    """
    Dataset importer for importing from ASpecD dataset format.

    For more details of the ASpecD dataset format, see the
    :class:`aspecd.io.AdfExporter` class.

    """

    def __init__(self, source=None):
        super().__init__(source=source)
        self.extension = ".adf"
        self._dataset_yaml_filename = "dataset.yaml"
        self._bin_dir = "binaryData"

    def _import(self):
        with tempfile.TemporaryDirectory() as tempdir:
            with zipfile.ZipFile(
                self.source + self.extension, "r"
            ) as zipped_file:
                zipped_file.extractall(path=tempdir)
                yaml = aspecd.utils.Yaml()
                yaml.binary_directory = os.path.join(tempdir, self._bin_dir)
                yaml.read_from(
                    os.path.join(tempdir, self._dataset_yaml_filename)
                )
                yaml.deserialise_numpy_arrays()
        self.dataset.from_dict(yaml.dict)


class AsdfExporter(DatasetExporter):
    """
    Dataset exporter for exporting to Advanced Scientific Data Format (ASDF).

    For more information on ASDF, see the
    `homepage of the asdf package <https://asdf.readthedocs.io/en/stable/>`_,
    and its `format specification <https://asdf-standard.readthedocs.io/>`_.

    """

    def __init__(self, target=None):
        super().__init__(target=target)
        self.extension = ".asdf"

    def _export(self):
        if not self.target:
            raise aspecd.exceptions.MissingTargetError

        dataset_dict = self.dataset.to_dict()
        dataset_dict["dataset_history"] = dataset_dict.pop("history")
        asdf_file = asdf.AsdfFile(dataset_dict)
        asdf_file.write_to(self.target + self.extension)


class AsdfImporter(DatasetImporter):
    """
    Dataset importer for importing from Advanced Scientific Data Format (ASDF).

    For more information on ASDF, see the
    `homepage of the asdf package <https://asdf.readthedocs.io/en/stable/>`_,
    and its `format specification <https://asdf-standard.readthedocs.io/>`_.

    """

    def __init__(self, source=None):
        super().__init__(source=source)
        self.extension = ".asdf"

    def _import(self):
        with asdf.open(
            self.source + self.extension, lazy_load=False, copy_arrays=True
        ) as asdf_file:
            dataset_dict = asdf_file.tree
            dataset_dict["history"] = dataset_dict.pop("dataset_history")
            self.dataset.from_dict(dataset_dict)


class TxtImporter(DatasetImporter):
    # noinspection PyUnresolvedReferences
    """
    Dataset importer for importing from plain text files (TXT).

    Plain text files have often the disadvantage of no accompanying metadata,
    therefore the use of plain text files for data storage is highly
    discouraged, besides other problems like inherent low resolution/accuracy
    or otherwise large file sizes.

    The main reason for this class to exist is that it provides a simple way
    to showcase ASpecD functionality reading from primitive data sources.
    Besides that, sometimes you will encounter plain text files.

    .. note::
        The importer relies on :func:`numpy.loadtxt` for reading text files.
        Hence, you can use any parameters understood by this function as
        keys in the :attr:`parameters` attribute. For handling decimal
        separators (non-standard behaviour), see below.

    If your data consist of two columns, the first will automatically be
    interpreted as the *x* axis. In all other cases, data will be read as is
    and no axes values explicitly written.


    Attributes
    ----------
    parameters : :class:`dict`
        Parameters controlling the import

        skiprows : :class:`int`
            Number of rows to skip in text file (*e.g.*, header lines)

        delimiter : :class:`str`
            The string used to separate values.

            Default: None (meaning: whitespace)

        comments : :class:`str` | :class:`list`
            Characters or list of characters indicating the start of a comment.

            Default: #

        separator : :class:`str`
            Character used as decimal separator.

            Default: None (meaning: dot)

        axis : :class:`int` | ``None``
            Column index to use for axis in case of 2D data.

            Often, when reading 2D text data, the first column contains
            the axis values. In case you don't have axis values in the
            data, set this parameter to ``None`` (in case of a YAML
            recipe, use ``Null`` instead).

            You can provide any (valid) column index here, starting with
            ``0`` for the first column. The column marked as axis is
            removed from the data array, but set as axis values.

            Default: 0

            .. versionadded:: 0.11


    .. admonition:: Decimal separators

        Handling decimal separators other than the dot is notoriously
        difficult, though often necessary. For this, the non-standard key
        ``separator`` has been introduced that is not supported by numpy
        itself. If you specify a character using this parameter, the file
        will be read as text and the character specified replaced by the
        dot. Only afterwards will :func:`numpy.loadtxt` be used with all the
        other parameters as usual.


    .. versionchanged:: 0.6.3
        Document more parameter keys; add handling of decimal separator.

    .. versionchanged:: 0.11
        For 2D data, the first column (index 0) is used as axis values by
        default.

    """

    def __init__(self, source=None):
        super().__init__(source=source)
        self.extension = ".txt"
        self.parameters["skiprows"] = 0
        self.parameters["delimiter"] = None
        self.parameters["comments"] = "#"
        self.parameters["separator"] = None
        self.parameters["axis"] = 0

    def _import(self):
        if "separator" in self.parameters:
            separator = self.parameters.pop("separator")
        else:
            separator = None
        if "axis" in self.parameters:
            axis = self.parameters.pop("axis")
        else:
            axis = 0
        if separator:
            with open(self.source, encoding="utf8") as file:
                contents = file.read()
            contents = contents.replace(separator, ".")
            # noinspection PyTypeChecker
            data = np.loadtxt(io.StringIO(contents), **self.parameters)
        else:
            data = np.loadtxt(self.source, **self.parameters)
        if data.ndim == 2 and axis is not None:
            self.dataset.data.axes[0].values = data[:, axis]
            data = np.delete(data, axis, 1)
            if data.shape[1] == 1:
                data = data[:, 0]
        self.dataset.data.data = data


class TxtExporter(DatasetExporter):
    """
    Dataset exporter for exporting to plain text files (TXT).

    Plain text files have often the disadvantage of no accompanying metadata,
    therefore the use of plain text files for data storage is highly
    discouraged, besides other problems like inherent low resolution/accuracy
    or otherwise large file sizes.

    .. warning::
        All metadata contained within a dataset (including the full history)
        are lost when exporting to plain text. Therefore, using this
        exporter will usually result in you loosing reproducibility. Hence,
        better think twice before using this exporter and use entirely on
        your own risk and only if you *really* know what you are doing (and
        why).

    The main reason for this class to exist is that sometimes there is a
    need to export data to a simple exchange format that can be shared with
    collaboration partners.

    .. note::
        The importer relies on :func:`numpy.savetxt` for writing text files.
        Hence, the same limitations apply, *e.g.* only working for 1D and 2D
        data, but not data with more than two dimensions.

    In case of 1D data, the resulting file will consist of two columns,
    with the first column consisting of the axis values and the second column
    containing the actual data. An example of the contents of such a file
    are given below:

    .. code-block::

        3.400000000000000000e+02 6.340967862812832978e-01
        3.410000000000000000e+02 3.424209074593306257e-01
        3.420000000000000000e+02 1.675116805484100357e-02

    In case of 2D data, the resulting file will contain the axes values in the
    first row/column respectively. Hence, the size of the matrix will be +1
    in both directions compared to the size of the actual data and the first
    element (top left) will always be zero (and shall be ignored). An
    example of the contents of such a file are given below:

    .. code-block::

        0.000000000000000000e+00 4.000000000000000000e+00 5.000000000000000000e+00
        3.400000000000000000e+02 6.340967862812832978e-01 5.979077980106655144e-01
        3.410000000000000000e+02 3.424209074593306257e-01 1.052868239245914328e-01
        3.420000000000000000e+02 1.675116805484100357e-02 9.050894282755458375e-01

    These two examples show immediately two of the problems of this file
    format: You are left to guess the quantity and unit of each the axes,
    and these files get quite big, as many decimal places are stored to
    not loose numerical resolution. With 50 characters per line for a 1D
    dataset (translating to at least one byte each), you end up with 50 kB
    for 1000 values.

    """

    def __init__(self, target=None):
        super().__init__(target=target)
        self.extension = ".txt"

    def _export(self):
        if not self.target:
            raise aspecd.exceptions.MissingTargetError

        if len(self.dataset.data.axes) == 2:
            data = np.asarray(
                [self.dataset.data.axes[0].values, self.dataset.data.data]
            ).T
        else:
            data = np.zeros(np.asarray(self.dataset.data.data.shape) + 1)
            data[1:, 0] = self.dataset.data.axes[0].values
            data[0, 1:] = self.dataset.data.axes[1].values
            data[1:, 1:] = self.dataset.data.data

        np.savetxt(self._sanitise_file_extension(self.target), data)

    def _sanitise_file_extension(self, target=None):
        return "".join([os.path.splitext(target)[0], self.extension])



"""Tests for history."""

import unittest
from datetime import datetime, timedelta

import aspecd.annotation
import aspecd.exceptions
import aspecd.history
import aspecd.system
import aspecd.utils
from aspecd import processing, dataset, analysis, plotting, table, io


class TestHistoryRecord(unittest.TestCase):
    def setUp(self):
        self.history_record = aspecd.history.HistoryRecord()

    def test_instantiate_class(self):
        pass

    def test_instantiate_class_with_package_name(self):
        aspecd.history.HistoryRecord(package="aspecd")

    def test_instantiate_class_with_package_name_sets_sysinfo(self):
        history = aspecd.history.HistoryRecord(package="numpy")
        self.assertTrue("numpy" in history.sysinfo.packages.keys())

    def test_has_date_property(self):
        self.assertTrue(hasattr(self.history_record, "date"))

    def test_date_is_datetime(self):
        self.assertTrue(isinstance(self.history_record.date, datetime))

    def test_date_is_current_date(self):
        now = datetime.today()
        # noinspection PyTypeChecker
        self.assertAlmostEqual(
            now, self.history_record.date, delta=timedelta(seconds=1)
        )

    def test_has_sysinfo_property(self):
        self.assertTrue(hasattr(self.history_record, "sysinfo"))

    def test_sysinfo_is_systeminfo(self):
        self.assertTrue(
            isinstance(self.history_record.sysinfo, aspecd.system.SystemInfo)
        )

    def test_has_to_dict_method(self):
        self.assertTrue(hasattr(self.history_record, "to_dict"))
        self.assertTrue(callable(self.history_record.to_dict))

    def test_has_from_dict_method(self):
        self.assertTrue(hasattr(self.history_record, "from_dict"))
        self.assertTrue(callable(self.history_record.from_dict))

    def test_from_dict_sets_date(self):
        orig_dict = self.history_record.to_dict()
        new_history_record = aspecd.history.HistoryRecord()
        new_history_record.from_dict(orig_dict)
        self.assertEqual(self.history_record.date, new_history_record.date)

    def test_from_dict_sets_sysinfo(self):
        orig_dict = self.history_record.to_dict()
        orig_dict["sysinfo"]["user"]["login"] = "foo"
        new_history_record = aspecd.history.HistoryRecord()
        new_history_record.from_dict(orig_dict)
        self.assertEqual(
            orig_dict["sysinfo"]["user"]["login"],
            new_history_record.sysinfo.user["login"],
        )

    def test_from_dict_sets_arbitrary_existing_attribute(self):
        orig_dict = self.history_record.to_dict()
        orig_dict["foo"] = "bar"
        new_history_record = aspecd.history.HistoryRecord()
        new_history_record.foo = ""
        new_history_record.from_dict(orig_dict)
        self.assertEqual(orig_dict["foo"], new_history_record.foo)

    def test_from_dict_ignores_arbitrary_non_existing_attribute(self):
        orig_dict = self.history_record.to_dict()
        orig_dict["foo"] = "bar"
        new_history_record = aspecd.history.HistoryRecord()
        new_history_record.from_dict(orig_dict)
        self.assertFalse(hasattr(new_history_record, "foo"))


class TestProcessingStepRecord(unittest.TestCase):
    def setUp(self):
        self.processing_step = processing.SingleProcessingStep()
        self.processing_record = aspecd.history.ProcessingStepRecord(
            self.processing_step
        )

    def test_instantiate_class(self):
        aspecd.history.ProcessingStepRecord()

    def test_instantiate_class_with_processing_step(self):
        aspecd.history.ProcessingStepRecord(self.processing_step)

    def test_instantiate_description_from_processing_step(self):
        self.processing_step.description = "Test"
        processing_record = aspecd.history.ProcessingStepRecord(
            self.processing_step
        )
        self.assertEqual(processing_record.description, "Test")

    def test_has_from_processing_step_method(self):
        self.assertTrue(
            hasattr(self.processing_record, "from_processing_step")
        )
        self.assertTrue(callable(self.processing_record.from_processing_step))

    def test_has_create_processing_step_method(self):
        self.assertTrue(
            hasattr(self.processing_record, "create_processing_step")
        )
        self.assertTrue(
            callable(self.processing_record.create_processing_step)
        )

    def test_create_processing_step_returns_processing_object(self):
        test_object = self.processing_record.create_processing_step()
        self.assertTrue(
            isinstance(test_object, processing.SingleProcessingStep)
        )

    def test_processing_object_has_correct_undoable_value(self):
        self.processing_record.undoable = True
        test_object = self.processing_record.create_processing_step()
        self.assertTrue(test_object.undoable, self.processing_record.undoable)

    def test_has_parameters_property(self):
        self.assertTrue(hasattr(self.processing_record, "parameters"))

    def test_processing_object_has_correct_parameters_value(self):
        self.processing_record.parameters["test"] = True
        test_object = self.processing_record.create_processing_step()
        self.assertEqual(test_object.parameters["test"], True)

    def test_has_description_property(self):
        self.assertTrue(hasattr(self.processing_record, "description"))

    def test_processing_object_has_correct_description_value(self):
        self.processing_record.description = "Test"
        test_object = self.processing_record.create_processing_step()
        self.assertEqual(test_object.description, "Test")

    def test_has_class_name_property(self):
        self.assertTrue(hasattr(self.processing_record, "class_name"))

    def test_has_comment_property(self):
        self.assertTrue(hasattr(self.processing_record, "comment"))

    def test_has_references_property(self):
        self.assertTrue(hasattr(self.processing_record, "references"))

    def test_processing_object_gets_correct_parameters_value(self):
        test_dictionary = dict(bla="blub", foo="bar")
        self.processing_step.parameters = test_dictionary
        self.processing_record = aspecd.history.ProcessingStepRecord(
            self.processing_step
        )
        test_object = self.processing_record.create_processing_step()
        self.assertEqual(test_object.parameters, test_dictionary)

    def test_processing_object_gets_correct_undoable_value(self):
        self.processing_record.undoable = True
        test_object = self.processing_record.create_processing_step()
        self.assertEqual(test_object.undoable, True)

    def test_processing_object_gets_correct_comment_value(self):
        test_comment = "Frobnicate the bizbaz"
        self.processing_step.comment = test_comment
        self.processing_record = aspecd.history.ProcessingStepRecord(
            self.processing_step
        )
        test_object = self.processing_record.create_processing_step()
        self.assertEqual(test_object.comment, test_comment)

    def test_processing_object_gets_correct_references_value(self):
        self.processing_step.references = ["foo"]
        self.processing_record = aspecd.history.ProcessingStepRecord(
            self.processing_step
        )
        test_object = self.processing_record.create_processing_step()
        self.assertEqual(test_object.references, ["foo"])

    def test_has_to_dict_method(self):
        self.assertTrue(hasattr(self.processing_record, "to_dict"))
        self.assertTrue(callable(self.processing_record.to_dict))

    def test_from_dict(self):
        orig_dict = self.processing_record.to_dict()
        orig_dict["comment"] = "foo"
        new_processing_record = aspecd.history.ProcessingStepRecord(
            self.processing_step
        )
        new_processing_record.from_dict(orig_dict)
        self.assertDictEqual(orig_dict, new_processing_record.to_dict())


class TestProcessingHistoryRecord(unittest.TestCase):
    def setUp(self):
        self.processing_step = processing.SingleProcessingStep()
        self.historyrecord = aspecd.history.ProcessingHistoryRecord(
            self.processing_step
        )

    def test_instantiate_class(self):
        pass

    def test_instantiate_class_with_processing_step(self):
        aspecd.history.ProcessingHistoryRecord(self.processing_step)

    def test_instantiate_class_with_package_name(self):
        aspecd.history.ProcessingHistoryRecord(
            processing_step=self.processing_step, package="numpy"
        )

    def test_instantiate_class_with_package_name_sets_sysinfo(self):
        processing_step = aspecd.history.ProcessingHistoryRecord(
            processing_step=self.processing_step, package="numpy"
        )
        self.assertTrue("numpy" in processing_step.sysinfo.packages.keys())

    def test_has_processing_property(self):
        self.assertTrue(hasattr(self.historyrecord, "processing"))

    def test_processing_is_processingsteprecord(self):
        self.assertTrue(
            isinstance(
                self.historyrecord.processing,
                aspecd.history.ProcessingStepRecord,
            )
        )

    def test_has_date_property(self):
        self.assertTrue(hasattr(self.historyrecord, "date"))

    def test_has_sysinfo_property(self):
        self.assertTrue(hasattr(self.historyrecord, "sysinfo"))

    def test_has_undoable_property(self):
        self.assertTrue(hasattr(self.historyrecord, "undoable"))

    def test_undoable_is_boolean(self):
        self.assertTrue(isinstance(self.historyrecord.undoable, bool))

    def test_has_replay_method(self):
        self.assertTrue(hasattr(self.historyrecord, "replay"))
        self.assertTrue(callable(self.historyrecord.replay))

    def test_replay(self):
        self.historyrecord.replay(dataset.Dataset())


class TestAnalysisStepRecord(unittest.TestCase):
    def setUp(self):
        self.analysis_step = analysis.AnalysisStep()
        self.analysis_record = aspecd.history.AnalysisStepRecord(
            self.analysis_step
        )

    def test_instantiate_class(self):
        aspecd.history.AnalysisStepRecord()

    def test_instantiate_class_with_analysis_step(self):
        aspecd.history.AnalysisStepRecord(self.analysis_step)

    def test_instantiate_description_from_analysis_step(self):
        self.analysis_step.description = "Test"
        analysis_record = aspecd.history.AnalysisStepRecord(
            self.analysis_step
        )
        self.assertEqual(analysis_record.description, "Test")

    def test_has_create_analysis_step_method(self):
        self.assertTrue(hasattr(self.analysis_record, "create_analysis_step"))
        self.assertTrue(callable(self.analysis_record.create_analysis_step))

    def test_has_from_processing_step_method(self):
        self.assertTrue(hasattr(self.analysis_record, "from_analysis_step"))
        self.assertTrue(callable(self.analysis_record.from_analysis_step))

    def test_create_analysis_step_returns_analysis_object(self):
        test_object = self.analysis_record.create_analysis_step()
        self.assertTrue(isinstance(test_object, analysis.AnalysisStep))

    def test_has_parameters_property(self):
        self.assertTrue(hasattr(self.analysis_record, "parameters"))

    def test_analysis_object_has_correct_parameters_value(self):
        self.analysis_record.parameters["test"] = True
        test_object = self.analysis_record.create_analysis_step()
        self.assertEqual(test_object.parameters["test"], True)

    def test_has_description_property(self):
        self.assertTrue(hasattr(self.analysis_record, "description"))

    def test_analysis_object_has_correct_description_value(self):
        self.analysis_record.description = "Test"
        test_object = self.analysis_record.create_analysis_step()
        self.assertEqual(test_object.description, "Test")

    def test_has_class_name_property(self):
        self.assertTrue(hasattr(self.analysis_record, "class_name"))

    def test_has_comment_property(self):
        self.assertTrue(hasattr(self.analysis_record, "comment"))

    def test_has_references_property(self):
        self.assertTrue(hasattr(self.analysis_record, "references"))

    def test_analysis_object_gets_correct_parameters_value(self):
        test_dictionary = dict(bla="blub", foo="bar")
        self.analysis_step.parameters = test_dictionary
        self.analysis_record = aspecd.history.AnalysisStepRecord(
            self.analysis_step
        )
        test_object = self.analysis_record.create_analysis_step()
        self.assertEqual(test_object.parameters, test_dictionary)

    def test_analysis_object_gets_correct_comment_value(self):
        test_comment = "Frobnicate the bizbaz"
        self.analysis_step.comment = test_comment
        self.analysis_record = aspecd.history.AnalysisStepRecord(
            self.analysis_step
        )
        test_object = self.analysis_record.create_analysis_step()
        self.assertEqual(test_object.comment, test_comment)

    def test_analysissteprecord_gets_correct_references_value(self):
        self.analysis_step.references = ["foo"]
        self.analysis_record = aspecd.history.AnalysisStepRecord(
            self.analysis_step
        )
        test_object = self.analysis_record.create_analysis_step()
        self.assertEqual(["foo"], test_object.references)

    def test_analysissteprecord_gets_result_from_analysisstep(self):
        test_result = ["foo"]
        self.analysis_step.result = test_result
        self.analysis_record = aspecd.history.AnalysisStepRecord(
            self.analysis_step
        )
        self.assertEqual(self.analysis_record.result, test_result)

    def test_analysisstep_from_record_has_no_result_value(self):
        test_result = ["foo"]
        self.analysis_step.result = test_result
        self.analysis_record = aspecd.history.AnalysisStepRecord(
            self.analysis_step
        )
        test_object = self.analysis_record.create_analysis_step()
        self.assertEqual(test_object.result, None)

    def test_has_to_dict_method(self):
        self.assertTrue(hasattr(self.analysis_record, "to_dict"))
        self.assertTrue(callable(self.analysis_record.to_dict))

    def test_from_dict(self):
        orig_dict = self.analysis_record.to_dict()
        orig_dict["comment"] = "foo"
        new_processing_record = aspecd.history.AnalysisStepRecord(
            self.analysis_step
        )
        new_processing_record.from_dict(orig_dict)
        self.assertDictEqual(orig_dict, new_processing_record.to_dict())


class TestSingleAnalysisStepRecord(unittest.TestCase):
    def setUp(self):
        self.analysis_step = analysis.SingleAnalysisStep()
        self.analysis_record = aspecd.history.SingleAnalysisStepRecord(
            self.analysis_step
        )

    def test_instantiate_class(self):
        pass


class TestAnalysisHistoryRecord(unittest.TestCase):
    def setUp(self):
        self.analysis_step = analysis.SingleAnalysisStep()
        self.historyrecord = aspecd.history.AnalysisHistoryRecord(
            self.analysis_step
        )

    def test_instantiate_class(self):
        pass

    def test_instantiate_class_with_package_name(self):
        aspecd.history.AnalysisHistoryRecord(
            analysis_step=self.analysis_step, package="numpy"
        )

    def test_instantiate_class_with_package_name_sets_sysinfo(self):
        analysis_step = aspecd.history.AnalysisHistoryRecord(
            analysis_step=self.analysis_step, package="numpy"
        )
        self.assertTrue("numpy" in analysis_step.sysinfo.packages.keys())

    def test_has_analysis_property(self):
        self.assertTrue(hasattr(self.historyrecord, "analysis"))

    def test_analysis_is_analysissteprecord(self):
        self.assertTrue(
            isinstance(
                self.historyrecord.analysis,
                aspecd.history.SingleAnalysisStepRecord,
            )
        )

    def test_has_replay_method(self):
        self.assertTrue(hasattr(self.historyrecord, "replay"))
        self.assertTrue(callable(self.historyrecord.replay))

    def test_replay(self):
        self.historyrecord.replay(dataset.Dataset())


class TestDatasetAnnotationRecord(unittest.TestCase):
    def setUp(self):
        self.annotation = aspecd.annotation.DatasetAnnotation()
        self.annotation_record = aspecd.history.AnnotationRecord(
            self.annotation
        )

    def test_instantiate_class(self):
        aspecd.annotation.DatasetAnnotation()

    def test_instantiate_class_with_annotation(self):
        aspecd.history.AnnotationRecord(self.annotation)

    def test_instantiate_content_from_annotation(self):
        self.annotation.content = {"foo": "bar"}
        annotation_record = aspecd.history.AnnotationRecord(self.annotation)
        self.assertEqual(annotation_record.content, self.annotation.content)

    def test_instantiate_class_name_from_annotation(self):
        annotation_record = aspecd.history.AnnotationRecord(self.annotation)
        self.assertEqual(
            annotation_record.class_name,
            aspecd.utils.full_class_name(self.annotation),
        )

    def test_has_from_annotation_method(self):
        self.assertTrue(hasattr(self.annotation_record, "from_annotation"))
        self.assertTrue(callable(self.annotation_record.from_annotation))

    def test_has_create_annotation_method(self):
        self.assertTrue(hasattr(self.annotation_record, "create_annotation"))
        self.assertTrue(callable(self.annotation_record.create_annotation))

    def test_create_annotation_returns_annotation_object(self):
        test_object = self.annotation_record.create_annotation()
        self.assertTrue(
            isinstance(test_object, aspecd.annotation.DatasetAnnotation)
        )

    def test_annotation_object_has_correct_contents_value(self):
        self.annotation_record.content = {"foo": "bar"}
        test_object = self.annotation_record.create_annotation()
        self.assertEqual(self.annotation_record.content, test_object.content)

    def test_has_to_dict_method(self):
        self.assertTrue(hasattr(self.annotation_record, "to_dict"))
        self.assertTrue(callable(self.annotation_record.to_dict))

    def test_from_dict(self):
        orig_dict = self.annotation_record.to_dict()
        orig_dict["content"]["comment"] = "foo"
        new_annotation_record = aspecd.history.AnnotationRecord()
        new_annotation_record.from_dict(orig_dict)
        self.assertDictEqual(
            orig_dict["content"], new_annotation_record.to_dict()["content"]
        )


class TestDatasetAnnotationHistoryRecord(unittest.TestCase):
    def setUp(self):
        self.annotation = aspecd.annotation.DatasetAnnotation()
        self.annotation_record = aspecd.history.AnnotationHistoryRecord(
            annotation=self.annotation
        )

    def test_instantiate_class(self):
        pass

    def test_instantiate_class_with_package_name(self):
        aspecd.history.AnnotationHistoryRecord(
            annotation=self.annotation, package="numpy"
        )

    def test_instantiate_class_with_package_name_sets_sysinfo(self):
        annotation_step = aspecd.history.AnnotationHistoryRecord(
            annotation=self.annotation, package="numpy"
        )
        self.assertTrue("numpy" in annotation_step.sysinfo.packages.keys())

    def test_has_annotation_property(self):
        self.assertTrue(hasattr(self.annotation_record, "annotation"))

    def test_annotation_is_annotation_record(self):
        self.assertIsInstance(
            self.annotation_record.annotation, aspecd.history.AnnotationRecord
        )


class TestPlotRecord(unittest.TestCase):
    def setUp(self):
        self.plot_record = aspecd.history.PlotRecord()

    def test_instantiate_class(self):
        pass

    def test_has_name_property(self):
        self.assertTrue(hasattr(self.plot_record, "class_name"))

    def test_has_parameters_property(self):
        self.assertTrue(hasattr(self.plot_record, "parameters"))

    def test_has_description_property(self):
        self.assertTrue(hasattr(self.plot_record, "description"))

    def test_has_filename_property(self):
        self.assertTrue(hasattr(self.plot_record, "filename"))

    def test_has_from_plotter_method(self):
        self.assertTrue(hasattr(self.plot_record, "from_plotter"))
        self.assertTrue(callable(self.plot_record.from_plotter))

    def test_instantiate_with_plotter(self):
        plotter = plotting.Plotter()
        aspecd.history.PlotRecord(plotter=plotter)

    def test_from_plotter_without_plotter_raises(self):
        with self.assertRaises(aspecd.exceptions.MissingPlotterError):
            self.plot_record.from_plotter()

    def test_from_plotter_sets_attributes(self):
        plotter = plotting.Plotter()
        plotter.filename = "test"
        caption = plotting.Caption()
        caption.title = "My fancy figure"
        caption.text = "Some more description"
        plotter.caption = caption
        plotter.label = "label"
        self.plot_record.from_plotter(plotter)
        self.assertEqual(plotter.name, self.plot_record.class_name)
        self.assertEqual(plotter.filename, self.plot_record.filename)
        self.assertEqual(plotter.parameters, self.plot_record.parameters)
        self.assertEqual(plotter.properties, self.plot_record.properties)
        self.assertEqual(plotter.description, self.plot_record.description)
        self.assertEqual(
            plotter.caption.title, self.plot_record.caption.title
        )
        self.assertEqual(plotter.label, self.plot_record.label)

    def test_instantiate_with_plotter_sets_attributes_from_plotter(self):
        plotter = plotting.Plotter()
        plotter.filename = "test"
        caption = plotting.Caption()
        caption.title = "My fancy figure"
        caption.text = "Some more description"
        plotter.caption = caption
        plot_record = aspecd.history.PlotRecord(plotter=plotter)
        self.assertEqual(plotter.name, plot_record.class_name)
        self.assertEqual(plotter.filename, plot_record.filename)
        self.assertEqual(plotter.parameters, plot_record.parameters)
        self.assertEqual(plotter.description, plot_record.description)
        self.assertEqual(plotter.caption.title, plot_record.caption.title)

    def test_has_to_dict_method(self):
        self.assertTrue(hasattr(self.plot_record, "to_dict"))
        self.assertTrue(callable(self.plot_record.to_dict))

    def test_from_dict(self):
        orig_dict = self.plot_record.to_dict()
        orig_dict["description"] = "foo"
        new_plot_record = aspecd.history.PlotRecord()
        new_plot_record.from_dict(orig_dict)
        self.assertDictEqual(orig_dict, new_plot_record.to_dict())


class TestSinglePlotRecord(unittest.TestCase):
    def setUp(self):
        self.plot_record = aspecd.history.SinglePlotRecord()

    def test_instantiate_class(self):
        pass

    def test_has_preprocessing_property(self):
        self.assertTrue(hasattr(self.plot_record, "preprocessing"))


class TestMultiPlotRecord(unittest.TestCase):
    def setUp(self):
        self.plot_record = aspecd.history.MultiPlotRecord()

    def test_instantiate_class(self):
        pass

    def test_has_datasets_property(self):
        self.assertTrue(hasattr(self.plot_record, "datasets"))


class TestTableRecord(unittest.TestCase):
    def setUp(self):
        self.table = aspecd.table.Table()
        self.table_record = aspecd.history.TableRecord(self.table)

    def test_instantiate_class(self):
        pass

    def test_instantiate_format_from_table(self):
        self.table.format = "dokuwiki"
        table_record = aspecd.history.TableRecord(self.table)
        self.assertEqual(table_record.format, self.table.format)

    def test_instantiate_class_name_from_table(self):
        table_record = aspecd.history.TableRecord(self.table)
        self.assertEqual(
            table_record.class_name, aspecd.utils.full_class_name(self.table)
        )

    def test_has_from_table_method(self):
        self.assertTrue(hasattr(self.table_record, "from_table"))
        self.assertTrue(callable(self.table_record.from_table))

    def test_has_create_table_method(self):
        self.assertTrue(hasattr(self.table_record, "create_table"))
        self.assertTrue(callable(self.table_record.create_table))

    def test_create_table_returns_table_object(self):
        test_object = self.table_record.create_table()
        self.assertTrue(isinstance(test_object, aspecd.table.Table))

    def test_table_object_has_correct_format_value(self):
        self.table.format = "dokuwiki"
        test_object = self.table_record.create_table()
        self.assertEqual(self.table_record.format, test_object.format)

    def test_has_to_dict_method(self):
        self.assertTrue(hasattr(self.table_record, "to_dict"))
        self.assertTrue(callable(self.table_record.to_dict))

    def test_from_dict(self):
        orig_dict = self.table_record.to_dict()
        orig_dict["format"] = "dokuwiki"
        new_table_record = aspecd.history.TableRecord()
        new_table_record.from_dict(orig_dict)
        self.assertDictEqual(orig_dict, new_table_record.to_dict())


class TestTableHistoryRecord(unittest.TestCase):
    def setUp(self):
        self.table = aspecd.table.Table()
        self.table_record = aspecd.history.TableHistoryRecord(
            table=self.table
        )

    def test_instantiate_class(self):
        pass

    def test_instantiate_class_with_package_name(self):
        aspecd.history.TableHistoryRecord(table=self.table, package="numpy")

    def test_instantiate_class_with_package_name_sets_sysinfo(self):
        table_ = aspecd.history.TableHistoryRecord(
            table=self.table, package="numpy"
        )
        self.assertTrue("numpy" in table_.sysinfo.packages.keys())

    def test_has_table_property(self):
        self.assertTrue(hasattr(self.table_record, "table"))

    def test_table_is_table_record(self):
        self.assertTrue(
            isinstance(self.table_record.table, aspecd.history.TableRecord)
        )


class TestDatasetExporterRecord(unittest.TestCase):
    def setUp(self):
        self.exporter = io.DatasetExporter()
        self.exporter_record = aspecd.history.DatasetExporterRecord(
            self.exporter
        )

    def test_instantiate_class(self):
        pass

    def test_instantiate_class_name_from_exporter(self):
        exporter_record = aspecd.history.DatasetExporterRecord(self.exporter)
        self.assertEqual(
            exporter_record.class_name,
            aspecd.utils.full_class_name(self.exporter),
        )

    def test_copy_attributes_from_exporter(self):
        self.exporter.target = "foobar"
        self.exporter.comment = "Lorem ipsum"
        exporter_record = aspecd.history.DatasetExporterRecord(self.exporter)
        self.assertEqual(exporter_record.target, self.exporter.target)
        self.assertEqual(exporter_record.comment, self.exporter.comment)

    def test_has_from_exporter_method(self):
        self.assertTrue(hasattr(self.exporter_record, "from_exporter"))
        self.assertTrue(callable(self.exporter_record.from_exporter))

    def test_has_create_exporter_method(self):
        self.assertTrue(hasattr(self.exporter_record, "create_exporter"))
        self.assertTrue(callable(self.exporter_record.create_exporter))

    def test_create_exporter_returns_exporter_object(self):
        test_object = self.exporter_record.create_exporter()
        self.assertTrue(isinstance(test_object, aspecd.io.DatasetExporter))

    def test_exporter_object_has_correct_attributes(self):
        self.exporter_record.target = "foo"
        self.exporter_record.comment = "Lorem ipsum"
        test_object = self.exporter_record.create_exporter()
        self.assertEqual(self.exporter_record.target, test_object.target)
        self.assertEqual(self.exporter_record.comment, test_object.comment)

    def test_has_to_dict_method(self):
        self.assertTrue(hasattr(self.exporter_record, "to_dict"))
        self.assertTrue(callable(self.exporter_record.to_dict))

    def test_from_dict(self):
        orig_dict = self.exporter_record.to_dict()
        orig_dict["target"] = "foo"
        new_exporter_record = aspecd.history.DatasetExporterRecord()
        new_exporter_record.from_dict(orig_dict)
        self.assertDictEqual(orig_dict, new_exporter_record.to_dict())

    def test_from_exporter_without_exporter_raises(self):
        with self.assertRaisesRegex(TypeError, "needs a DatasetExporter"):
            self.exporter_record.from_exporter()


class TestDatasetExporterHistoryRecord(unittest.TestCase):
    def setUp(self):
        self.exporter = io.DatasetExporter()
        self.exporter_record = aspecd.history.DatasetExporterHistoryRecord(
            exporter=self.exporter
        )

    def test_instantiate_class(self):
        pass

    def test_instantiate_class_with_package_name(self):
        aspecd.history.DatasetExporterHistoryRecord(
            exporter=self.exporter, package="numpy"
        )

    def test_instantiate_class_with_package_name_sets_sysinfo(self):
        exporter = aspecd.history.DatasetExporterHistoryRecord(
            exporter=self.exporter, package="numpy"
        )
        self.assertTrue("numpy" in exporter.sysinfo.packages.keys())

    def test_has_exporter_property(self):
        self.assertTrue(hasattr(self.exporter_record, "exporter"))

    def test_exporter_is_exporter_record(self):
        self.assertTrue(
            isinstance(
                self.exporter_record.exporter,
                aspecd.history.DatasetExporterRecord,
            )
        )

    def test_exporter_has_correct_attributes(self):
        self.exporter.target = "foo"
        exporter = aspecd.history.DatasetExporterHistoryRecord(
            exporter=self.exporter
        )
        self.assertEqual(self.exporter.target, exporter.exporter.target)



"""Metadata: Information on numeric data stored in a structured way.

Metadata are one key concept of the ASpecD framework, and they come in
different flavours. Perhaps the easiest to grasp is metadata that accompany
measurements---and are often stored separately from the data in metadata
files. Other types of metadata are those of processing steps or
representations. This module is only concerned with the first kind of
metadata, information accompanying data and usually recorded at the same
time as the numeric data.

Each :class:`aspecd.dataset.Dataset` contains a
:attr:`aspecd.dataset.Dataset.metadata` attribute that is in turn an object
of class :class:`aspecd.metadata.DatasetMetadata`. This latter object is
composed of different metadata objects each inheriting from
:class:`aspecd.metadata.Metadata`. Upon import of a dataset, the importer
class needs to make sure that as many metadata as possible are read and
imported to the dataset as well.

Generally speaking, metadata can be thought of as key--value stores that
might be hierarchically structured and thus cascaded. Nevertheless, classes
have some advantages over using simple dictionaries, as there are certain
operations that are common to some or all types of metadata.


Metadata classes
================

The classes implemented in this module can be grouped into general metadata
classes, concrete metadata classes, and metadata classes for datasets. Each
will be described shortly below.


General metadata classes
------------------------

The most basic class is :class:`aspecd.metadata.PhysicalQuantity` storing
all relevant information about a physical quantity in an easily accessible
way, eventually allowing to test for commensurable quantities and
converting between units.

Next is :class:`aspecd.metadata.Metadata` as a generic class for all
metadata containers. All other classes storing metadata, particularly those
storing metadata accompanying measurements and therefore ending up in the
metadata of a :class:`aspecd.dataset.Dataset`, should inherit from this class.


Concrete metadata classes
-------------------------

Currently, three classes for actual metadata of experimental datasets and
one class for calculated datasets are contained in the ASpecD framework,
namely :class:`aspecd.metadata.Measurement` for storing
general information about a given measurement,
:class:`aspecd.metadata.Sample` for all information regarding the sample
investigated, and :class:`aspecd.metadata.TemperatureControl` for
information about the temperature control (including whether temperature has
been actively controlled at all during the measurement) for experimental
datasets and :class:`aspecd.metadata.Calculation` for storing details about
the calculation underlying the (numeric) data for calculated datasets.


Metadata classes for datasets
-----------------------------

The attribute `metadata` in the :class:`aspecd.dataset.ExperimentalDataset`
is of type :class:`aspecd.metadata.ExperimentalDatasetMetadata` and
contains the three metadata classes for experimental datasets named above.
Derived packages should extend this class accordingly.

Similarly, the attribute `metadata` in the
:class:`aspecd.dataset.CalculatedDataset` is of type
:class:`aspecd.metadata.CalculatedDatasetMetadata` and contains the respective
metadata class named above. Derived packages should extend this class
accordingly wherever necessary.


Converting metadata from and to dictionaries
============================================

All classes inheriting from :class:`aspecd.metadata.Metadata` provide a
method :meth:`from_dict` allowing to set the attributes of the objects. This
allows for easy use with metadata read from a file into a `dict`.

Similiarly, all classes inheriting from :class:`aspecd.metadata.Metadata` as
well as :class:`aspecd.metadata.PhysicalQuantity` provide a method
:meth:`to_dict` that returns a dictionary of all public attributes of the
respective object. This allows to write metadata to a file.


Mapping metadata read from external sources
===========================================

Generally, the representation and structure of metadata within the dataset
of the ASpecD framework and each application derived from it is separate
from the way the very same metadata are organised in files written mostly
during data acquisition. To map the structure obtained by reading a
metadata file to the internal representation within the dataset, as given
by the :class:`aspecd.metadata.ExperimentalDatasetMetadata` class,
there exists a generic mapper class :class:`aspecd.metadata.MetadataMapper`.
This way, you can separate the representations of metadata and support
mapping for different versions of metadata files.


.. note::

    As mappings can become quite complicated and specifying lists of
    mappings for the :class:`aspecd.metadata.MetadataMapper` by hand can
    become quite tedious, you can specify metadata mapping recipes in YAML
    files in a rather simple syntax. See the documentation of the
    :class:`aspecd.metadata.MetadataMapper` class and its
    :meth:`aspecd.metadata.MetadataMapper.create_mappings` method for details.

    This method and the underlying ideas are heavily based on concepts
    and code developed by J. Popp for use within the `trEPR Python package
    <https://docs.trepr.de/>`_.


Metadata in packages based on the ASpecD framework
==================================================

The dataset as unit of (numerical) data and metadata is a key concept of the
ASpecD framework and a necessary prerequisite for a semantic understanding
within the routines. Every measurement (or calculation) produces (raw) data
that are useless without additional information, such as experimental
parameters. This additional information is termed "metadata" within the
ASpecD framework.

Additionally to combining numerical data and metadata, a dataset provides a
common structure, unifying the different file formats used as source for
both, data and metadata. Hence, the actual data format does not matter,
greatly facilitating dealing with data from different sources (and even
different kinds of data).

Therefore, if you develop a new package based on the ASpecD framework,
one of the first and most important steps is to create a (hierarchical)
metadata structure for your datasets. This requires a thorough understanding
of the spectroscopic method you develop the package for and most probably
several years of practical experience in the lab. Good sources of
inspiration are the vendor file formats usually storing instrument
parameters and alike in some form. If you are lucky, you can actually access
this information. If not, you may need to store these metadata in an
additional external file that gets written manually during data recording.

Some basic metadata that are rarely contained within vendor file formats,
as they concern the actual sample measured, as well as metadata for
calculated datasets can be found in the
:class:`aspecd.metadata.ExperimentalDatasetMetadata` and
:class:`aspecd.metadata.CalculatedDatasetMetadata`.


Module documentation
====================

"""

import datetime
import logging

import aspecd.exceptions
import aspecd.utils


logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class PhysicalQuantity(aspecd.utils.ToDictMixin):
    """
    Class for storing all relevant informations about a physical quantity.

    A physical quantity, Q, consists always of a value, {Q} and a
    corresponding unit, [Q], hence:

    Q = {Q} [Q] .

    See, e.g., the "IUPAC Green Book" for further details.

    To get a string representation of a physical quantity, i.e., its value
    and unit separated by a single space, simply use :meth:`str()`.

    To set value (and unit) from a string, either use :meth:`from_string` or
    supply the string as argument while instantiating the object. Make sure
    the value part of the string to be convertible to float. Otherwise,
    a :class:`ValueError` will be raised.

    Attributes
    ----------
    unit : :class:`str`
        symbol of the unit of the corresponding value

        SI units are preferred
    dimension : :class:`str`
        dimension of the corresponding value

        useful for (automatic) conversions
    name : :class:`str`
        name of the physical quantity in a given context

    Parameters
    ----------
    string : :class:`str`
        String containing value and unit, separated by whitespace

        Will be used to set value and unit correspondingly.

        If no second element separated by whitespace is present,
        only :attr:`value` will be set.
    value : `float`
        Numerical value
    unit : :class:`str`
        String containing the unit of the corresponding value.

        SI units are preferred, and their abbreviations should be used.

    Raises
    ------
    ValueError
        Raised if value is not a float

    """

    def __init__(self, string=None, value=None, unit=None):
        super().__init__()
        self._value = 0.0
        self.unit = ""
        self.dimension = ""
        self.name = ""
        if string:
            self._set_value_unit_from_string(string)
        if value:
            self.value = value
        if unit:
            self.unit = unit
        self._include_in_to_dict = ["value"]

    def __str__(self):
        """String representation of object."""
        if self.value == 0.0 and not self.unit:
            string = ""
        else:
            string = " ".join([str(self.value), self.unit])
        return string

    @property
    def value(self):
        """
        Get or set the value of a physical quantity.

        A value is always a float.

        Raises
        ------
        ValueError
            Raised if value is not a (scalar) float

        """
        return self._value

    @value.setter
    def value(self, value=None):
        if not isinstance(value, type(self._value)):
            raise TypeError
        self._value = value

    def _set_value_unit_from_string(self, string):
        parts = string.split(maxsplit=1)
        self.value = float(parts[0].strip())
        if len(parts) > 1:
            self.unit = parts[1].strip()

    def commensurable(self, physical_quantity=None):
        """
        Check whether two physical quantities are commensurable.

        There are two criteria for physical quantities to be commensurable.
        Either they have the same unit, or they have the same dimension. In
        the latter case, a unit conversion is generally possible.

        Parameters
        ----------
        physical_quantity : :obj:`aspecd.metadata.PhysicalQuantity`
            physical quantity to test commensurability with

        Returns
        -------
        commensurable : :class:`bool`
            True if both physical quantities have the same unit or
            dimension, False otherwise

        """
        commensurable = False
        if (
            self.unit
            and hasattr(physical_quantity, "unit")
            and physical_quantity.unit
            and self.unit == physical_quantity.unit
        ):
            commensurable = True
        if (
            self.dimension
            and hasattr(physical_quantity, "dimension")
            and physical_quantity.dimension
            and self.dimension == physical_quantity.dimension
        ):
            commensurable = True
        return commensurable

    def from_string(self, string):
        """
        Set value and unit from string.

        Parameters
        ----------
        string : :class:`str`
            String containing value and unit, separated by whitespace

            Will be used to set value and unit correspondingly.

            If no second element separated by whitespace is present,
            only :attr:`value` will be set.

            If an empty string is provided, value and unit are cleared.

        """
        if string and not string.lower() == "n/a":
            self._set_value_unit_from_string(string)
        else:
            self.value = 0.0
            self.unit = ""

    def from_dict(self, dict_=None):
        """
        Set properties from dictionary.

        Only parameters in the dictionary that are valid properties of the
        class are set accordingly.

        Parameters
        ----------
        dict_ : :class:`dict`
            Dictionary containing properties to set

        """
        for key in dict_:
            if hasattr(self, key):
                setattr(self, key, dict_[key])


class Metadata(aspecd.utils.ToDictMixin):
    """
    General metadata class.

    Metadata can be set from dict upon initialisation.

    Metadata can be converted to dict via
    :meth:`aspecd.utils.ToDictMixin.to_dict()`.

    """

    def __init__(self, dict_=None):
        """
        Instantiate Metadata object.

        Parameters
        ----------
        dict_ : :class:`dict`
            Dictionary containing properties to set

        """
        super().__init__()
        if dict_:
            self.from_dict(dict_)

    def from_dict(self, dict_=None):
        """
        Set properties from dictionary, e.g., from metadata.

        Only parameters in the dictionary that are valid properties of the
        class are set accordingly.

        Keys in the dictionary are converted to lower case and spaces
        converted to underscores to fit the naming scheme of attributes.

        If a property is of class :class:`aspecd.metadata.PhysicalQuantity`,
        it is set accordingly.

        Parameters
        ----------
        dict_ : :class:`dict`
            Dictionary containing properties to set

        """
        for key in dict_:
            attribute_name = key.replace(" ", "_").lower()
            if hasattr(self, attribute_name):
                if isinstance(
                    getattr(self, attribute_name), PhysicalQuantity
                ):
                    if isinstance(dict_[key], dict):
                        getattr(self, attribute_name).from_dict(dict_[key])
                    else:
                        getattr(self, attribute_name).from_string(dict_[key])
                else:
                    setattr(self, attribute_name, dict_[key])


class TemperatureControl(Metadata):
    """
    Temperature control is very often found in spectroscopy.

    This class provides general means of storing relevant parameters for
    temperature control.

    Attributes
    ----------
    temperature : :class:`aspecd.metadata.PhysicalQuantity`
        value and unit of the temperature set
    controller : :class:`str`
        type and name of the temperature controller used

    """

    def __init__(self, dict_=None, temperature=""):
        """
        Instantiate TemperatureControl object.

        If both, temperature and dictionary containing a temperature key are
        given, the dictionary key will overwrite the temperature parameter.

        Parameters
        ----------
        dict_ : :class:`dict`
            Dictionary containing properties to set
        temperature : :class:`str`
            Value and unit of temperature

        """
        self.temperature = PhysicalQuantity(temperature)
        self.controller = ""
        super().__init__(dict_=dict_)

    @property
    def controlled(self):
        """
        Has temperature been actively controlled during measurement?

        Read-only property automatically set when setting a temperature value.

        """
        return self.temperature.value

    def from_dict(self, dict_=None):
        """
        Set properties from dictionary, e.g., from metadata.

        Only parameters in the dictionary that are valid properties of the
        class are set accordingly.

        If "controlled" is set to False in the dictionary, the temperature
        value and unit will be cleared.

        The value of the temperature key needs to be a string.

        Parameters
        ----------
        dict_ : :class:`dict`
            Dictionary with keys corresponding to properties of the class.

        """
        # Get value for key "controlled" if it exists, and delete it from dict_
        controlled_in_dict = "controlled" in dict_
        if controlled_in_dict:
            controlled = dict_["controlled"]
            del dict_["controlled"]
        else:
            controlled = False
        super().from_dict(dict_)
        # Checking "controlled" needs to be done after assigning other keys
        if controlled_in_dict and not controlled:
            self.temperature.from_string("")


class Measurement(Metadata):
    """
    General information available for each type of measurement.

    Attributes
    ----------
    start : `datetime`
        Date and time of start of measurement
    end : `datetime`
        Date and time of end of measurement
    purpose: :class:`str`
        Purpose of measurement, often quite helpful to know
    operator : :class:`str`
        Name of the operator performing the measurement
        Beware of the implications for privacy protection
    labbook_entry : :class:`str`
        Identifier for lab book entry (usually either LOI or URL)

    Parameters
    ----------
    dict_ : :class:`dict`
        Dictionary containing fields corresponding to attributes of the class

    """

    def __init__(self, dict_=None):
        self.start = None
        self.end = None
        self.purpose = ""
        self.operator = ""
        self.labbook_entry = ""
        super().__init__(dict_=dict_)

    def from_dict(self, dict_=None):
        """
        Set properties from dictionary, e.g., from metadata.

        Only parameters in the dictionary that are valid properties of the
        class are set accordingly.

        For the "start" and "end" items, there are two different conventions
        available how the dictionary can be structured. Either those fields
        are dictionaries themselves, with fields "date" and "time"
        accordingly, such as::

            {"start": {"date": "yyyy-mm-dd", "time": "HH:MM:SS"},
             "end": {"date": "yyyy-mm-dd", "time": "HH:MM:SS"}}

        Alternatively, those fields can be strings containing a
        representation of both, date and time::

            {"start": "yyyy-mm-dd HH:MM:SS", "end": "yyyy-mm-dd HH:MM:SS"}

        Use whichever is more appropriate for you.

        Parameters
        ----------
        dict_ : :class:`dict`
            Dictionary with keys corresponding to properties of the class.

        """
        for key in ("start", "end"):
            if key in dict_:
                if isinstance(dict_[key], dict):
                    self._set_datetime_from_dict(key, dict_.pop(key))
                else:
                    self._set_datetime_from_string(key, dict_.pop(key))
        super().from_dict(dict_=dict_)

    def _set_datetime_from_dict(self, key="", dict_=None):
        """
        Set start and end properties from dictionary containing values.

        The dictionary will usually be obtained from reading a metadata file.
        Date and time formats are restricted, following "%Y-%m-%d" for date
        and "%H:%M:%S" for time.

        Note that datetime objects are immutable.

        Parameters
        ----------
        key : :class:`str`
            Property of class to set
        dict_ : :class:`dict`
            Dictionary with fields "date" and "time"

        """
        if isinstance(dict_["date"], datetime.date):
            dict_["date"] = dict_["date"].isoformat()
        datetime_string = " ".join([dict_["date"], dict_["time"]])
        self._set_datetime_from_string(key=key, string=datetime_string)

    def _set_datetime_from_string(self, key="", string=""):
        """
        Set start and end properties from string containing values.

        The string will usually be obtained from reading a metadata file.
        Date and time formats are restricted, following "%Y-%m-%d" for date
        and "%H:%M:%S" for time, separated by a single space.

        Note that datetime objects are immutable.

        Parameters
        ----------
        key : :class:`str`
            Property of class to set
        string : :class:`str`
            String containing date and time in format described above

        """
        if string:
            fmt = "%Y-%m-%d %H:%M:%S"
            setattr(self, key, datetime.datetime.strptime(string, fmt))


class Sample(Metadata):
    """
    Information on the sample measured.

    Attributes
    ----------
    name : :class:`str`
        Short name of the sample
    id : :class:`str` or `int`
        Unique identifier of the sample
    loi : :class:`str`
        Lab Object Identifier (LOI) for the sample

    Parameters
    ----------
    dict_ : :class:`dict`
        Dictionary containing fields corresponding to attributes of the class

    """

    def __init__(self, dict_=None):
        self.name = ""
        self.id = None  # pylint: disable=invalid-name
        self.loi = ""
        super().__init__(dict_=dict_)


class Calculation(Metadata):
    """
    Information on the calculation.

    Attributes
    ----------
    type : :class:`str`
        Type of the calculation -- usually the full class name
    parameters : :class:`dict`
        Parameters of the calculation

    Parameters
    ----------
    dict_ : :class:`dict`
        Dictionary containing fields corresponding to attributes of the class

    """

    def __init__(self, dict_=None):
        self.type = ""
        self.parameters = {}
        super().__init__(dict_=dict_)


class Device(Metadata):
    """
    Information on the device contributing device data.

    The dataset concept (see :class:`aspecd.dataset.Dataset`) rests on the
    assumption that there is one particular set of data that can be
    regarded as the actual or primary data of the dataset. However,
    in many cases, parallel to these actual data, other data are recorded
    as well, be it readouts from monitors or alike.

    This class contains the metadata of the corresponding devices whose
    data are of type :class:`aspecd.dataset.DeviceData`. That class
    contains an attribute :attr:`aspecd.dataset.DeviceData.metadata`.

    .. note::
        You will usually need to implement derived classes for concrete
        devices, as this class only contains a minimum set of attributes.


    Attributes
    ----------
    label : :class:`str`
        Label of the device

    Parameters
    ----------
    dict_ : :class:`dict`
        Dictionary containing fields corresponding to attributes of the class


    .. versionadded:: 0.9

    """

    def __init__(self, dict_=None):
        self.label = ""
        super().__init__(dict_=dict_)


class DatasetMetadata(aspecd.utils.ToDictMixin):
    """
    Metadata for dataset.

    This class contains the minimal set of metadata for a dataset.

    Metadata of actual datasets should extend this class by adding
    properties that are themselves classes inheriting from
    :class:`aspecd.metadata.Metadata`.

    Metadata can be converted to dict via
    :meth:`aspecd.utils.ToDictMixin.to_dict()`, e.g., for generating
    reports using templates and template engines.

    """

    # pylint: disable=useless-super-delegation
    def __init__(self):
        super().__init__()

    def from_dict(self, dict_=None):
        """
        Set properties from dictionary, e.g., from metadata.

        Only parameters in the dictionary that are valid properties of the
        class are set accordingly.

        Keys in the dictionary are converted to lower case and spaces
        converted to underscores to fit the naming scheme of attributes.

        Parameters
        ----------
        dict_ : :class:`dict`
            Dictionary with metadata.

            Each key of this dictionary corresponds to a class attribute and
            is in itself a dictionary with the correct set of attributes for
            the particular class.

        """
        for key in dict_:
            attr = key.replace(" ", "_").lower()
            if hasattr(self, attr):
                getattr(self, attr).from_dict(dict_[key])


class ExperimentalDatasetMetadata(DatasetMetadata):
    """
    Metadata for an experimental dataset.

    This class contains the minimal set of metadata for an experimental
    dataset, i.e., :class:`aspecd.dataset.ExperimentalDataset`.

    Metadata of actual datasets should extend this class by adding
    properties that are themselves classes inheriting from
    :class:`aspecd.metadata.Metadata`.

    Metadata can be converted to dict via
    :meth:`aspecd.utils.ToDictMixin.to_dict()`, e.g., for generating
    reports using templates and template engines.

    Attributes
    ----------
    measurement : :class:`aspecd.metadata.Measurement`
        Metadata of measurement
    sample : :class:`aspecd.metadata.Sample`
        Metadata of sample
    temperature_control : :class:`aspecd.metadata.TemperatureControl`
        Metadata of temperature control

    """

    def __init__(self):
        super().__init__()
        self.measurement = Measurement()
        self.sample = Sample()
        self.temperature_control = TemperatureControl()


class CalculatedDatasetMetadata(DatasetMetadata):
    """
    Metadata for a dataset with calculated data.

    This class contains the minimal set of metadata for a dataset consisting
    of calculated data, i.e., :class:`aspecd.dataset.CalculatedDataset`.

    Metadata of actual datasets should extend this class by adding
    properties that are themselves classes inheriting from
    :class:`aspecd.metadata.Metadata`.

    Metadata can be converted to dict via
    :meth:`aspecd.utils.ToDictMixin.to_dict()`, e.g., for generating
    reports using templates and template engines.

    Attributes
    ----------
    calculation : :class:`aspecd.metadata.Calculation`
        Metadata of calculation underlying the numeric data

    """

    def __init__(self):
        super().__init__()
        self.calculation = aspecd.metadata.Calculation()


class MetadataMapper:
    """
    Mapper for metadata.

    Allows to convert a dictionary containing metadata read, e.g., from a
    metadata file to a dictionary that corresponds to the internal
    structure of the metadata in a dataset stored in
    :class:`aspecd.metadata.ExperimentalDatasetMetadata`.

    If all you need is to convert the dictionary keys to proper variable
    names conforming to the naming scheme proposed by `PEP 8
    <https://pep8.org/>`_, you may simply use the method
    :meth:`keys_to_variable_names`.

    Tasks that can be currently performed to map a dictionary to the
    internal structure of the metadata representation in a dataset contain
    renaming of keys via :meth:`rename_key` and combining items via
    :meth:`combine_items` as well as copying keys via :meth:`copy_key` and
    moving items via :meth:`move_item`.

    Rather than performing the mappings by hand, calling these methods
    repeatedly, you may use a mapping table contained in the
    :attr:`mappings` attribute. If you pre-define such mapping tables,
    you can easily apply different mappings depending on the version of
    your original metadata structure. Once you assigned the appropriate
    mapping table to the :attr:`mappings` attribute, simply call
    :meth:`map`. If everything turns out well, this should map your
    metadata contained in :attr:`metadata` according to the mapping table
    contained in :meth:`mappings`. Finally, you may want to assing this
    converted data structure to your dataset's metadata attribute,
    using :meth:`ExperimentalDatasetMetadata.from_dict()`.

    As it is often tedious to manually create the entries of the mapping
    table residing in :attr:`mappings`, you can use mapping recipes stored
    in YAML files together with the :meth:`create_mappings` method. For
    details of the structure of the mapping recipe YAML files, see the
    documentation of the :meth:`create_mappings` method. Note that you need
    to specify the filename for the mapping recipe used in
    :attr:`recipe_filename` as well as the version of the metadata file
    format in :attr:`version` to get this to work. The :meth:`create_mappings`
    method and the underlying ideas are heavily based on concepts and code
    developed by J. Popp for use within the `trEPR Python package
    <https://docs.trepr.de/>`_.

    .. note::

        The mapping recipes should be stored within the package, and as
        accessing files from within packages should *not* be done using
        regular fiile-system paths, but rather the respective functionality
        of the pkgutil package used. Therefore, internally,
        the :func:`aspecd.utils.get_package_data` function is used. In case
        of using the :class:`MetadataMapper` class from a package derived
        from ASpecD, prefix the name of the recipe file with the package,
        followed by the '@' character. As an example, if you would want to
        use the recipe 'mappings.yaml' from within the package 'trepr',
        you would need to specify ``trepr@mappings.yaml`` as
        :attr`recipe_filename`.


    Attributes
    ----------
    metadata : :class:`dict`
        Dictionary containing the metadata that are converted in place

    mappings : :class:`list`
        Tasks to perform to map dictionary

        Each task is a list containing three entries:

        (1) an optional key of a "sub-dictionary" to operate on
        (2) the action to carry out
        (3) a list containing the necessary parameters to carry out the action

        For examples, see the documentation of the :meth:`map` method.

    version : :class:`str`
        Version of the metadata to map

        Particularly important when you use :meth:`create_mappings` to
        create the mappings from mapping recipes stored in a YAML file.

    recipe_filename : :class:`str`
        Name of the YAML file containing the mapping recipes

        Needs to be specified when you use :meth:`create_mappings` to
        create the mappings from mapping recipes stored in a YAML file.

    Examples
    --------
    To actually use the mapper, you will usually create a file (in YAML
    format) containing the mappings. For details how this file may look
    like, see the :meth:`create_mappings` method. Suppose you have saved
    your mappings to the file ``mappings.yaml``, with different mappings for
    the different versions of your formats. In this case, using the mapper
    may look similar to the following:

    .. code-block::

        mapper = aspecd.metadata.MetadataMapper()
        mapper.version = version_string
        mapper.metadata = dict_to_be_mapped
        mapper.recipe_filename = 'mappings.yaml'
        mapper.map()
        modified_dict = mapper.metadata

    As you can see, the ``modified_dict`` contains the dictionary where the
    mappings from ``mappings.yaml`` have been applied to.


    .. versionchanged:: 0.6
        Recipe files now retrieved from package data now via
        :func:`aspecd.utils.get_package_data`

    """

    def __init__(self):
        self.metadata = {}
        self.mappings = []
        self.version = ""
        self.recipe_filename = ""
        self._mapping_recipes = None
        self._mapping_recipe = None

    def rename_key(self, old_key="", new_key=""):
        """
        Rename key in dictionary.

        Note that this method does *not* preserve the order of keys in an
        ordered dictionary.

        Parameters
        ----------
        old_key : str
            Name of original key that shall be renamed
        new_key : str
            New name of key to be renamed

        """
        self._rename_key_in_dict(old_key, new_key, self.metadata)

    @staticmethod
    def _rename_key_in_dict(old_key="", new_key="", dict_=None):
        if old_key in dict_:
            dict_[new_key] = dict_.pop(old_key)
        else:
            logger.debug('Key "%s" not found', old_key)
        return dict_

    def combine_items(self, old_keys=None, new_key="", pattern=""):
        """
        Combine two items in a dictionary.

        Parameters
        ----------
        old_keys : list
            Keys that should be combined
        new_key : str
            Name of new key
        pattern : str
            Pattern to use to join the keys together.

            Defaults to the empty string.

        """
        self._combine_items_in_dict(old_keys, new_key, pattern, self.metadata)

    @staticmethod
    def _combine_items_in_dict(
        old_keys=None, new_key="", pattern="", dict_=None
    ):
        value_tmp = []
        for key in old_keys:
            value_tmp.append(dict_.pop(key))
        dict_[new_key] = pattern.join(value_tmp)
        return dict_

    def keys_to_variable_names(self):
        """
        Convert keys in :attr:`metadata` to proper variable names.

        Variable names in Python should be all lower case, with words
        joined by underscores.

        Due to recursively traversing through the :attr:`metadata`
        dictionary, conversion is performed for (near) arbitrary depth.
        """
        self.metadata = self._traverse_keys_to_variable_names(self.metadata)

    def _traverse_keys_to_variable_names(self, dict_=None):
        for key, value in dict_.copy().items():
            if isinstance(value, dict):
                dict_[key] = self._traverse_keys_to_variable_names(value)
            new_key = key.replace(" ", "_").lower()
            dict_[new_key] = dict_.pop(key)
        return dict_

    def copy_key(self, old_key="", new_key=""):
        """
        Copy key in dictionary to new key.

        This method is particularly useful in cases where keys need to be
        combined using :meth:`combine_keys`, but where one of the keys
        should be combined several times with another key.

        Parameters
        ----------
        old_key : str
            Name of original key that shall be copied
        new_key : str
            Name of new key to be added

        """
        self._copy_key_in_dict(
            old_key=old_key, new_key=new_key, dict_=self.metadata
        )

    @staticmethod
    def _copy_key_in_dict(old_key="", new_key="", dict_=None):
        dict_[new_key] = dict_[old_key]

    def move_item(
        self,
        key="",
        source_dict_name="",
        target_dict_name="",
        create_target_dict=False,
    ):
        """
        Move item (i.e., key-value pair) between dictionaries.

        .. note::
            If the target dictionary does not exist, usually the method
            will *not* create it and raise an appropriate exception.
            However, if explicitly told to create the target dictionary,
            it will do so. This is to prevent accidental typos from
            messing up with the dictionary and resulting in hard to track bugs.

        Parameters
        ----------
        key : str
            Name of the key of the corresponding item to move
        source_dict_name : str
            Name of the dict the item should be moved from
        target_dict_name : str
            Name of the dict the item should be moved to
        create_target_dict : bool
            Whether to create target dictionary if it doesn't exist

        """
        self._move_item_in_dict(
            key=key,
            source_dict_name=source_dict_name,
            target_dict_name=target_dict_name,
            create_target_dict=create_target_dict,
            dict_=self.metadata,
        )

    @staticmethod
    def _move_item_in_dict(
        key="",
        source_dict_name="",
        target_dict_name="",
        create_target_dict=False,
        dict_=None,
    ):
        if create_target_dict and target_dict_name not in dict_:
            dict_[target_dict_name] = {}
        if source_dict_name:
            source_dict = dict_[source_dict_name]
        else:
            source_dict = dict_
        if key in source_dict:
            dict_[target_dict_name][key] = source_dict.pop(key)
        else:
            logger.debug('Key "%s" not found in "%s"', key, source_dict_name)

    def map(self):
        """
        Map according to mappings in :attr:`mappings`.

        Each mapping is defined as a list containing optionally a key for a
        sub-dictionary as first element, the method to be performed as
        second element, and the parameters for this method as third element.

        An example for a mapping may look like this::

            mapping = [['', 'rename_key', ['old', 'new']]]

        This would rename the key ``old`` in :attr:`metadata` to ``new``.

        To do the same for a key in a "sub-dictionary", you may provide a
        mapping similar to the following::

            mapping = [['test', 'rename_key', ['old', 'new']]]

        This would rename the key ``old`` in the dictionary ``test`` in
        :attr:`metadata` to ``new``. The same pattern optionally specifying
        a dictionary to operate on can be applied to all the other mappings
        detailed below.

        Similarly, you can join two items to a new item. In this case,
        a mapping may look like this::

            mapping = [['', 'combine_items', [['key1', 'key2'], 'new']]]

        This would join the values corresponding to the two keys ``key1`` and
        ``key2`` and assign them to the new key ``new``. If you would like
        to join the values with a particular string, this can be done as well::

            mapping = [['', 'combine_items', [['key1', 'key2'], 'new', ' ']]]

        Here, the two values will be joined using a space.

        Sometimes you want to combine keys, but need one of the two keys
        several times. Hence, you would like to *first* copy this key to
        another one. This can be done in the following way::

            mapping = [['', 'copy_key', ['old', 'new']]]

        And finally, there are cases where you want to move an item from
        one dictionary to another. This can be done using the following
        mapping::

            mapping = [['', 'move_item', ['key', 'source', 'target']]]

        Here, "source" and "target" are the names of the respective
        dictionaries the item should be moved between. If the target
        dictionary does not exist, by default, the method will raise an
        exception. If, however, you decide to exactly know what you do,
        you can pass an additional parameter explicitly telling the method
        to create the target dictionary::

            mapping = [['', 'move_item', ['key', 'source', 'target', True]]]

        In this particular case, however, you are solely responsible for
        any typos when specifying the name of the target dictionary,
        as this will most probably mess up your dictionaries and result in
        hard to track bugs.

        """
        if self.recipe_filename:
            self.create_mappings()
        for mapping in self.mappings:
            if mapping[0]:
                method = getattr(self, "".join(["_", mapping[1], "_in_dict"]))
                if mapping[0] in self.metadata:
                    method(*mapping[2], dict_=self.metadata[mapping[0]])
                else:
                    logger.debug('Key "%s" not found', mapping[0])
            else:
                method = getattr(self, mapping[1])
                method(*mapping[2])

    def create_mappings(self):
        """
        Create mappings from mapping recipe stored in YAML file.

        Mapping recipes are stored in an external file (currently a YAML
        file whose filename is stored in :attr:`recipe_filename`) in their
        own format described hereafter. From this file, the recipes are read
        and converted into mappings in the :attr:`mappings` attribute.

        Based on the version number of the format the metadata from an external
        source are stored in, the correct recipe is selected.

        Following is an example of a YAML file containing recipes. Each map can
        contain several types of mappings and the latter can contain several
        entries:

        .. code-block:: yaml

            ---

            format:
              type: metadata mapper
              version: '0.1'

            map 1:
              metadata file versions:
                - 0.1.6
                - 0.1.5
              combine items:
                - old keys: ['Date start', 'Time start']
                  new key: start
                  pattern: ' '
                  in dict: GENERAL
              rename key:
                - old key: GENERAL
                  new key: measurement
                  in dict:

            map 2:
              metadata file versions:
                - 0.1.4
              copy key:
                - old key: Date
                  new key: Date end
                  in dict: GENERAL
              move item:
                - key: model
                  source dict: measurement
                  target dict: spectrometer
                - key: Runs
                  source dict: measurement
                  target dict: experiment
                  create target: True

        Unknown mappings are silently ignored. The difference between the
        two entries in ``move item`` is that in the latter case, the target
        dictionary will be created. Be careful with this option, as typos
        introduced in your mapping recipe will lead to hard-to-debug
        behaviour of your application. See :meth:`move_item` for details.

        .. important::

            If you have version numbers with only one dot, you need to
            explicitly mark this as string in YAML, as otherwise, it will
            automatically be converted into a float and hence your version
            lookup will fail.

        Generally, the YAML file should be pretty self-explanatory. For
        details of the different mappings, see the documentation of the
        respective methods of the class, namely :meth:`combine_items`,
        :meth:`rename_key`, :meth:`copy_key`, and :meth:`move_item`.

        .. important::

            The sequence of operations can sometimes be crucial. They are
            called as follows: "copy key" -> "combine items" -> "rename key"
            -> "remove items"

        Note that you can name the mappings called here ``map 1`` and ``map 2``
        as you like. Use descriptive names wherever possible.

        A hint on the filenames for metadata recipe YAML files: Use
        descriptive names containing the format of the metadata files. For
        info files, something like ``infofile_metadata_mappings.yaml`` may
        be reasonable.

        This method and the underlying ideas are heavily based on concepts
        and code developed by J. Popp for use within the `trEPR Python package
        <https://docs.trepr.de/>`_.

        """
        if not self.recipe_filename:
            raise aspecd.exceptions.MissingFilenameError(
                message="You need to provide a recipe filename in "
                "self.recipe_filename"
            )
        self._load_mapping_recipe_from_file()
        self._choose_mapping_recipe()
        self._add_mappings_from_recipe()

    def _load_mapping_recipe_from_file(self):
        yaml_file = aspecd.utils.Yaml()
        # noinspection PyTypeChecker
        yaml_file.read_stream(
            aspecd.utils.get_package_data(self.recipe_filename)
        )
        self._mapping_recipes = yaml_file.dict

    def _choose_mapping_recipe(self):
        for key in self._mapping_recipes.keys():
            if key != "format":
                if (
                    self.version
                    in self._mapping_recipes[key]["metadata file versions"]
                ):
                    self._mapping_recipe = self._mapping_recipes[key]
        if not self._mapping_recipe:
            raise aspecd.exceptions.MissingRecipeError(
                message="No matching recipe found."
            )

    def _add_mappings_from_recipe(self):
        if "copy key" in self._mapping_recipe.keys():
            for i in range(len(self._mapping_recipe["copy key"])):
                mapping = [
                    self._mapping_recipe["copy key"][i]["in dict"],
                    "copy_key",
                    [
                        self._mapping_recipe["copy key"][i]["old key"],
                        self._mapping_recipe["copy key"][i]["new key"],
                    ],
                ]
                self.mappings.append(mapping)
        if "combine items" in self._mapping_recipe.keys():
            for i in range(len(self._mapping_recipe["combine items"])):
                mapping = [
                    self._mapping_recipe["combine items"][i]["in dict"],
                    "combine_items",
                    [
                        self._mapping_recipe["combine items"][i]["old keys"],
                        self._mapping_recipe["combine items"][i]["new key"],
                        self._mapping_recipe["combine items"][i]["pattern"],
                    ],
                ]
                self.mappings.append(mapping)
        if "rename key" in self._mapping_recipe.keys():
            for i in range(len(self._mapping_recipe["rename key"])):
                mapping = [
                    self._mapping_recipe["rename key"][i]["in dict"],
                    "rename_key",
                    [
                        self._mapping_recipe["rename key"][i]["old key"],
                        self._mapping_recipe["rename key"][i]["new key"],
                    ],
                ]
                self.mappings.append(mapping)
        if "move item" in self._mapping_recipe.keys():
            for i in range(len(self._mapping_recipe["move item"])):
                sub_mapping = [
                    self._mapping_recipe["move item"][i]["key"],
                    self._mapping_recipe["move item"][i]["source dict"],
                    self._mapping_recipe["move item"][i]["target dict"],
                ]
                if "create target" in self._mapping_recipe["move item"][i]:
                    sub_mapping.append(True)
                mapping = ["", "move_item", sub_mapping]
                self.mappings.append(mapping)
